/* 
 * File:   XrdDPMCommon.hh
 * Author: Fabrizio Furano (CERN IT/GT), David Smith
 * 
 * Common functions
 *
 * Created on February 2, 2012, 1:57 PM
 */

#ifndef XRDDPMCOMMON_HH
#define	XRDDPMCOMMON_HH

#include <string.h>
#include <time.h>
#include <vector>

#include "XrdSec/XrdSecEntity.hh"
#include "XrdOuc/XrdOucString.hh"
#include "XrdOuc/XrdOucEnv.hh"
#include "XrdOuc/XrdOucName2Name.hh"

#include "dpns_api.h"
#include "dpm_api.h"

#define XRDDPM_IS_PUT 1
#define XRDDPM_ID_URL 2

/**********************************/
/*           classes              */
/**********************************/

class DFThreadStore {
public:
	~DFThreadStore();
	char	errbuf[4096];
	int	errbuf_refcount;

friend class DFThreadSpecific;
private:
	DFThreadStore();
};

class DFThreadSpecific {
public:
	DFThreadSpecific();
	~DFThreadSpecific();
	DFThreadStore*	GetStore();
	char*	(*th_getenv)(const char*);
	int	(*th_setenv)(const char*, const char*, int);
	bool	has_thenv;
	XrdOucString	orig_mechstr;

private:
	static void	destruct(void *p);
	pthread_key_t	key;
};

class DpmError {
public:
	DpmError(const char *fn);
	~DpmError();
	void	assign(const char *info);
	void	save(const char *cmd=0);
	const char*	GenerateBrief();
	const char*	GenerateDetail();
	int	GetSerrno() const;

private:
	char*	apibuf;
	size_t	apibuflen;
	std::vector<char>	oldbuf;
	XrdOucString	fn;
	XrdOucString	cmd;
	XrdOucString	extra;
	XrdOucString	savedapimsg;
	XrdOucString	rstr_b;
	XrdOucString	rstr_d;
	int	my_serrno;
};

class DpmIdentityException : public std::exception {
public:
	DpmIdentityException(int errnum, const char *detail);
	~DpmIdentityException() throw();
	virtual const char*	what() const throw();
	int	ErrNum() const;
	const char*	ErrStr() const;
	const char*	DetailStr() const;

private:
	int	errnum;
	XrdOucString	errmsg;
	XrdOucString	detailmsg;
};

class DpmIdentityConfigOptions {
public:
	XrdOucString	principal;
	std::vector<XrdOucString>	fqans;
	std::vector<XrdOucString>	validvo;
};


class DpmIdentity {
public:
	DpmIdentity(XrdOucEnv *Env, DpmIdentityConfigOptions &config);
	DpmIdentity(XrdOucEnv *Env);
	DpmIdentity();
	~DpmIdentity();
	void	SetAuth();
	void	UnsetAuth();
	const XrdOucString&	Dn() const;
	const XrdOucString&	Groups() const;
	bool	IsSecEntID() const;
	static bool	usesPresetID(XrdOucEnv *Env, const XrdSecEntity *Entity=0);
	static bool	badPresetID(DpmIdentityConfigOptions &config,
                XrdOucString &err);

private:
	DpmIdentity(const DpmIdentity&);
	DpmIdentity&	operator=(const DpmIdentity&);
	void	parse_grps();
	void	parse_secent(const XrdSecEntity *secEntity);
	void	check_validvo(DpmIdentityConfigOptions &config);
	XrdOucString	m_name;
	std::vector<XrdOucString>	m_vorgs;
	const char**	m_grp_array;
	std::vector<XrdOucString>	vs;
	XrdOucString	m_endors_raw;
	bool	UsesSecEntForID;
};

/**********************************/
/*          functions             */
/**********************************/

XrdOucString DecodeString(XrdOucString in);

// Encode an array of bytes to base64
char *Tobase64(const unsigned char *input, int length);


// Calculates the url hash, from the given information
// The hash is calculated from:
// - xrd_fn, the filename in the xroot request
// - surl of the file (needed elsewhere to close it)
// - dpmdhost
// - sfn
// - dpm request token 
// - flags, set of bit mapped flags
// - DN of the client  
// - voms string
// - creation time of the url
// - validity time before and after creation time          
// - nonce string
// - key
// - key length
//
char *calcHash(
	const char *xrd_fn,
	const char *surl,
	const char *dpmdhost,
	const char *sfn,
	const char *dpmtk,
	unsigned int flags,
	const char *dn,
	const char *vomsnfo,
	time_t tim,
	int tim_grace,
	const char *nonce,
	unsigned const char *key,
	size_t keylen ); 

int compareHash(
        const char *h1,
        const char *h2);

const char *getfnamestartfromsfn(const char *sfn);

inline const char *SafeCStr(const XrdOucString &in) {
        const char *p = in.c_str();
        if (!p)
                return "";  
        return p;
}

void InitLocalHostNameList(std::vector<XrdOucString> &names);

const char *LoadKeyFromFile(unsigned char **dat, size_t *dsize);

class DpmRedirConfigOptions {
public:
        XrdOucString    defaultPrefix;
        std::vector<std::pair<XrdOucString, XrdOucString > >    pathPrefixes;
	DpmIdentityConfigOptions        IdentConfig;
	std::vector<XrdOucString>	AuthLibRestrict;
	XrdOucName2Name*	theN2N;
	XrdOucString	lroot_param;
	std::vector<XrdOucString>	N2NCheckPrefixes;
};

XrdOucString CanonicalisePath(const char *s, int trailing_slash);

XrdOucString TranslatePath(
                DpmRedirConfigOptions &config,
                const char *in);

#endif	/* XRDDPMCOMMON_HH */
