/*
 * Copyright (C) 2012 by CERN/IT/GT/DMS. All rights reserved.
 *
 * Author: Fabrizio Furano, David Smith <David.Smith@cern.ch>
*/

#ifndef __DPM_FINDER__
#define __DPM_FINDER__

#include <pthread.h>
#include <vector>
#include <memory>
#include <utility>

#include "XrdDPM/XrdDPMCommon.hh"
#include "XrdCms/XrdCmsClient.hh"
#include "XrdCms/XrdCmsFinder.hh"
#include "XrdOuc/XrdOucString.hh"
#include "XrdOuc/XrdOucEnv.hh"
#include "XrdOuc/XrdOucErrInfo.hh"
#include "XrdSys/XrdSysPthread.hh"
#include "XrdAcc/XrdAccAuthorize.hh"

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include "dpm_api.h"

/**********************************/
/* classes internal to DPM Finder */
/**********************************/

class DpmFinderConfigOptions {
public:
	DpmFinderConfigOptions();

	unsigned short	xrd_server_port;
	bool	mkpath_bool;
	std::vector<sockaddr_storage>	mmReqHosts;

	time_t	reqput_lifetime;
	char	reqput_ftype;
	char	reqput_stoken[CA_MAXDPMTOKENLEN+1];
	u_signed64	reqput_reqsize;
	time_t	reqget_lifetime;
	char	reqget_ftype;
	char	reqget_stoken[CA_MAXDPMTOKENLEN+1];

	DpmRedirConfigOptions		RedirConfig;
	std::vector<unsigned char>	key;
	int	gracetime;

	XrdOucString	authlib;
	XrdOucString	authparm;
	bool	authorize_bool;

	XrdOucString    N2N_Lib;
	XrdOucString    N2N_Parms;
};

class DpmFileRequestException : public std::exception {
public:
	DpmFileRequestException(int status, const char *brief=0,
		const char *detail=0);
	~DpmFileRequestException() throw();
	virtual const char*	what() const throw();
	int	DpmStatus() const;
	int	ErrCode() const;
	const char*	ErrStr() const;
	const char*	DetailStr() const;

private:
	int	status;
	XrdOucString	errmsg;
	XrdOucString	detailmsg;
};

class DpmFileRequestOptions {
public:
	DpmFileRequestOptions(bool isput, XrdOucEnv *env,
		const DpmFinderConfigOptions &defaults);
	~DpmFileRequestOptions();
	bool	operator==(const DpmFileRequestOptions &rhs) const;
	bool	operator!=(const DpmFileRequestOptions &rhs) const;

	bool	isPut;
	time_t	lifetime;
	char	ftype;
	char	stoken[CA_MAXDPMTOKENLEN+1];
	u_signed64	reqsize;
	bool	isMkpath;
};

class DpmFileRequest {
public:
	DpmFileRequest(const XrdOucString &path, int flags,
			DpmFileRequestOptions &ReqOpts);
	~DpmFileRequest();
	int	LastStatus() const;
	void	UpdateStatus();
	void	Abort();
	int	BackoffWaitTime();
	const XrdOucString&	Host() const;
	const XrdOucString&	Sfn() const;
	const XrdOucString&	Path() const;
	const XrdOucString	RToken() const;
	const DpmFileRequestOptions&	ROpts() const;

private:
	void	init();

	class put_helper {
	public:
		put_helper(struct dpm_putfilereq *req,
			struct dpm_putfilestatus **statuses, int *reps);
		~put_helper();
	private:
		put_helper(const put_helper&);
		put_helper&	operator=(const put_helper&);
		struct dpm_putfilereq*	req;
		struct dpm_putfilestatus**	statuses;
		int*	reps;
	};

	class get_helper {
	public:
		get_helper(struct dpm_getfilereq *req,
			struct dpm_getfilestatus **statuses, int *reps);
		~get_helper();

	private:
		get_helper(const get_helper&);
		get_helper&	operator=(const get_helper&);
		struct dpm_getfilereq*	req;
		struct dpm_getfilestatus**	statuses;
		int*	reps;
	};

	void	do_first_put();
	void	do_putstatus();
	void	do_first_get();
	void	do_getstatus();
	void	parse_turl(const char *turl);
	bool	isAccessiblePutOngoing();

	int	waittime;
	bool	isFirst;
	bool	withOverwrite;
	XrdOucString	spath;
	int	flags;
	DpmFileRequestOptions	ReqOpts;
	char	r_token[CA_MAXDPMTOKENLEN+1];
	int	ReqStatus;
	XrdOucString	r_host, r_sfn;
	int	MkpathState;
};

class DFQueueKey {
public:
	DFQueueKey(const XrdOucString &path, const XrdOucString &Dn,
		const XrdOucString &Groups,
		const DpmFileRequestOptions &ReqOpts);
	~DFQueueKey();

	bool	operator==(const DFQueueKey&) const;
	bool	operator!=(const DFQueueKey&) const;
                        
private:
	XrdOucString	path;
	XrdOucString	Dn;
	XrdOucString	Groups;
	DpmFileRequestOptions	ReqOpts;
};

class DFQueueItem {
public:
	DFQueueItem(const DFQueueKey &key);
	~DFQueueItem();

	DpmFileRequest*	GetReq();
	void	SetReq(DpmFileRequest *req);
	time_t	CreateTime() const;
	void	SetSkipDelay(time_t delay);

friend class DFQueueAnchor;
private:
	DFQueueItem*	next;
	std::auto_ptr<DFQueueKey>	key;
	time_t	create_time;
	time_t	skip_until;
	DpmFileRequest*	req;
	int	tag;

	// object not copyable because of ownership of req
	DFQueueItem(const DFQueueItem&);
	DFQueueItem&	operator=(const DFQueueItem&);
};

class DFQueueAnchor {
public:
	DFQueueAnchor();
	~DFQueueAnchor();

	DFQueueItem*	Remove(const DFQueueKey &key);
	void	Insert(DFQueueItem *item);

private:
	DFQueueItem*	first;
	XrdSysMutex	myMutex;
	int	seq;
};


/******************************************************************************/
/*                             D P M   F i n d e r                            */
/******************************************************************************/

class XrdDPMFinder : public XrdCmsClient
{
public:
	int	Configure(const char *cfn, char *Args, XrdOucEnv *EnvInfo);

	int	Forward(XrdOucErrInfo &Resp, const char *cmd,
			const char *arg1=0,  const char *arg2=0,
			XrdOucEnv  *Env1=0,  XrdOucEnv  *Env2=0);

	int	Locate(XrdOucErrInfo &Resp, const char *path, int flags,
			XrdOucEnv *Info=0);

	int	Space(XrdOucErrInfo &Resp, const char *path, XrdOucEnv *Info=0);

		XrdDPMFinder(XrdCmsClient *cmsClient, XrdSysLogger *lp,
			int whoami=0, int Port=0);
		~XrdDPMFinder();

private:
	int	setupN2N(XrdSysError &Eroute, const char *cfn);
	int	setupAuth(XrdSysLogger *lp, const char *cfn);
	int	DoFileAccessRequest(XrdOucErrInfo &Resp, const char *xrd_fn,
		const XrdOucString &FullPath, int flags, bool isPut,
		const DpmIdentity &ident, XrdOucEnv *Info);
	bool	IsMetaManagerDiscover(XrdOucEnv *Info, int flags,
		const char *path, const char *user);

	DpmFinderConfigOptions	DefaultConfig;
	XrdAccAuthorize*	Authorization;
	bool	AuthSecondary;
	XrdCmsClient*	defaultCmsClient;
};
#endif
