/* 
 * File:   XrdDPMOss.hh
 * Author: Fabrizio Furano, David Smith (CERN IT/GT)
 * 
 * An Oss plugin that interacts with a DPM
 *
 * Created on December 21, 2011, 10:25 AM
 */

#ifndef XRDDPMOSS_HH
#define	XRDDPMOSS_HH

#include <memory>

#include "XrdDPM/XrdDPMCommon.hh"
#include "XrdOss/XrdOssApi.hh"

class XrdDPMOssDir : public XrdOssDir {
public:
	int Close(long long *retsz = 0);
	int Opendir(const char *, XrdOucEnv &);
	int Readdir(char *buff, int blen);

	// Constructor and destructor

	XrdDPMOssDir(const char *tid);
	~XrdDPMOssDir();

private:
	std::auto_ptr<DpmIdentity> identP;
	DpmRedirConfigOptions *RedirConfig;
	dpns_DIR *dpns_dirp;
};

class XrdDPMOssFile : public XrdOssFile {
public:

	// The following two are virtual functions to allow for upcasting derivations
	// of this implementation
	//
	virtual int Close(long long *retsz = 0);
	virtual int Open(const char *, int, mode_t, XrdOucEnv &);
	virtual int Fchmod(mode_t mode);

	// Constructor and destructor

	XrdDPMOssFile(const char *tid, char *fnpfx): XrdOssFile(tid), mydpmtk(0), myfilename(0), myfnpfx(fnpfx), myOflag(0) {
	}

	virtual ~XrdDPMOssFile();

private:
	// The dpm request token, used for a get/put
	char *mydpmtk;
	// the filename for the original dpm request
	char *myfilename;
	// A prefix to internally add the filename to
	char *myfnpfx;
	// The original open flags
	int myOflag;
	// identity to use for eventual close
	std::auto_ptr<DpmIdentity> identP;
};

class XrdDPMOss : public XrdOss {
public:
	
	virtual XrdOssDF *newDir(const char *tident)
                       {return (XrdOssDF *)new XrdDPMOssDir(tident);}
	virtual XrdOssDF *newFile(const char *tident)
                       {return (XrdOssDF *)new XrdDPMOssFile(tident, myfnpfx);}

	virtual int Chmod(const char *, mode_t mode, XrdOucEnv *eP = 0);
	virtual int Create(const char *, const char *, mode_t, XrdOucEnv &, int opts = 0);
	virtual int Init(XrdSysLogger *, const char *);
	virtual int Mkdir(const char *, mode_t mode, int mkpath = 0, XrdOucEnv *eP = 0);

	virtual int Reloc(const char *, const char *, const char *, const char *x = 0) {
		return -ENOTSUP;
	}
	virtual int Remdir(const char *, int Opts = 0, XrdOucEnv *eP = 0);
	virtual int Rename(const char *, const char *,
		XrdOucEnv *eP1 = 0, XrdOucEnv *eP2 = 0);
	virtual int Stat(const char *, struct stat *, int opts = 0, XrdOucEnv *eP = 0);

	virtual int StatFS(const char *path, char *buff, int &blen, XrdOucEnv *eP = 0) {
		return -ENOTSUP;
	}

	virtual int StatLS(XrdOucEnv &env, const char *cgrp, char *buff, int &blen) {
		return -ENOTSUP;
	}

	virtual int StatXA(const char *path, char *buff, int &blen, XrdOucEnv *eP = 0) {
		return -ENOTSUP;
	}

	virtual int StatXP(const char *path, unsigned long long &attr,
		XrdOucEnv *eP = 0) {
		return -ENOTSUP;
	}
	virtual int Truncate(const char *, unsigned long long, XrdOucEnv *eP = 0) {
		return -ENOTSUP;
	}
	virtual int Unlink(const char *fn, int Opts = 0, XrdOucEnv *eP = 0);

	virtual int Stats(char *bp, int bl) {
		return 0;
	}

	virtual int StatVS(XrdOssVSInfo *sP, const char *sname = 0, int updt = 0) {
		return -ENOTSUP;
	}

	virtual int Lfn2Pfn(const char *Path, char *buff, int blen);

	virtual
	const char *Lfn2Pfn(const char *Path, char *buff, int blen, int &rc);

	XrdDPMOss(XrdOss *OssRoute, char *fnpfx): XrdOss(), sharedkey(0) {
		// Object to route the calls to after the DPM-related 
		// parameter massaging
		ossroute = OssRoute;
		
		// Prefix to use for the incoming requests, e.g. /dpm/cern.ch
		myfnpfx = 0;
		
		if (fnpfx) myfnpfx = strdup(fnpfx);
	}

	virtual ~XrdDPMOss() {
		if (ossroute) delete ossroute;
		if (myfnpfx) free(myfnpfx);
	}
	
	
	char *sharedkey;
private:
	XrdOss *ossroute;
	char *myfnpfx;
};



#endif	/* XRDDPMOSS_HH */



/******************************************************************************/
/*           S t o r a g e   S y s t e m   I n s t a n t i a t o r            */
/******************************************************************************/

// This function is called to obtain an instance of a configured XrdOss object.
// It is passed the object that would have been used as the storage system.
// The object is not initialized (i.e., Init() has not yet been called).
// This allows one to easily wrap the native implementation or to completely 
// replace it, as needed. The name of the config file and any parameters
// specified after the path on the ofs.osslib directive are also passed (note
// that if no parameters exist, parms may be null).

extern "C" {
	XrdOss *XrdOssGetStorageSystem(XrdOss *native_oss,
		XrdSysLogger *Logger,
		const char *config_fn,
		const char *parms);
}
