/* 
 * File:   XrdDPMCommon.cc
 * Author: Fabrizio Furano (CERN IT/GT), David Smith
 * 
 * Common functions
 *
 * Created on February 2, 2012, 1:57 PM
 */

#include "XrdDPM/XrdDPMCommon.hh"

#include <string.h>
#include <openssl/hmac.h>
#include <openssl/bio.h>
#include <openssl/buffer.h>
#include <openssl/err.h>

#include <memory>
#include <vector>
#include <algorithm>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <fcntl.h>
#include <dlfcn.h>
#include <stdlib.h>

#include "Cthread_api.h"
#include "serrno.h"

extern "C" int Cdomainname(char *, int);

/******************************************************************************/
/*                                 classes                                    */
/******************************************************************************/

/******************************************************************************/
/*                          D F T h r e a d S t o r e                         */
/******************************************************************************/

DFThreadStore::~DFThreadStore() {}

DFThreadStore::DFThreadStore() : errbuf_refcount(0) {}

/******************************************************************************/
/*                       D F T h r e a d S p e c i f i c                      */
/******************************************************************************/

DFThreadSpecific::~DFThreadSpecific() {
	if (pthread_key_delete(key)) {
		std::cerr << "pthread_key_delete failed" << std::endl;
		abort();
	}
}

DFThreadStore *DFThreadSpecific::GetStore() {
	DFThreadStore *tp = (DFThreadStore *)pthread_getspecific(key);
	if (!tp) {
		std::auto_ptr<DFThreadStore> inst(new DFThreadStore);
		if (pthread_setspecific(key, inst.get())) {
			std::cerr << "pthread_setspecific failed" << std::endl;
			abort();
		}
		tp = inst.release();
	}
	return tp;
}

DFThreadSpecific::DFThreadSpecific() {
	const char *p;
	Cthread_init();
	umask(0);

	typedef char* (*thget_t)(const char*);
	typedef int (*thset_t)(const char*, const char*, int);

	th_getenv = reinterpret_cast<thget_t>(reinterpret_cast<size_t>
		(dlsym(RTLD_DEFAULT,"Cthread_getenv")));
	th_setenv = reinterpret_cast<thset_t>(reinterpret_cast<size_t>
		(dlsym(RTLD_DEFAULT,"Cthread_setenv")));

	if (th_getenv && th_setenv) {
		has_thenv = 1;
		if ((p = th_getenv("CSEC_MECH"))) {
			orig_mechstr.assign(p,0);
		} else {
			orig_mechstr = "GSI ID";
		}
	} else {
		has_thenv = 0;
		setenv("CSEC_MECH", "ID", 1);
	}
	if (pthread_key_create(&key, destruct)) {
		std::cerr << "pthread_key_create failed" << std::endl;
		abort();
	}
}

void DFThreadSpecific::destruct(void *p) {
	DFThreadStore *tp = (DFThreadStore *)p;
	delete tp;
	ERR_remove_state(0);
}

/******************************************************************************/
/*                               D p m E r r o r                              */
/******************************************************************************/

extern DFThreadSpecific dpm_sf;

DpmError::DpmError(const char *fn) : fn(fn) {
	DFThreadStore *ts = dpm_sf.GetStore();
	apibuf = ts->errbuf;
	apibuflen = sizeof(ts->errbuf);
	if (ts->errbuf_refcount++) {
		for(size_t i=0;i<apibuflen;++i)
			oldbuf.push_back(apibuf[i]);
	}
	dpns_seterrbuf(apibuf, apibuflen);
	dpm_seterrbuf(apibuf, apibuflen);
	apibuf[0] = '\0';
	serrno = my_serrno = 0;
}

DpmError::~DpmError() {
	DFThreadStore *ts = dpm_sf.GetStore();
	if (--ts->errbuf_refcount) {
		if (oldbuf.size() != apibuflen) {
			std::cerr <<
				"unexpected oldbuf size in ~DpmError()" <<
				std::endl;
			abort();
		}
		memcpy(apibuf, &oldbuf[0], apibuflen);
	} else {
		dpns_seterrbuf(0,0);
		dpm_seterrbuf(0,0);
	}
}

void DpmError::assign(const char *info) { extra.assign(info,0); }

void DpmError::save(const char *c) {
	my_serrno = serrno;
	if (c)
		cmd.assign(c,0);
	while(*apibuf && apibuf[strlen(apibuf)-1]=='\n')
		apibuf[strlen(apibuf)-1]='\0';
	savedapimsg.assign(apibuf, 0);
}

int DpmError::GetSerrno() const { return my_serrno; }

const char *DpmError::GenerateDetail() {
	rstr_d += "Error in " + fn;

	if (my_serrno) {
		rstr_d += ", serrno=";
		rstr_d += my_serrno;
		rstr_d += " [";
		rstr_d += sstrerror(my_serrno);
		rstr_d += "]";
	}

	if (savedapimsg.length()) {
		rstr_d += " <" + savedapimsg + ">";
	}
	while(extra.length() && extra.endswith('\n'))
		extra.erasefromend(1);
	if (extra.length()) {
		rstr_d += " (" + extra;
		rstr_d += ")";
	}
	if (cmd.length()) {
		rstr_d += " while calling " + cmd;
	}
	return SafeCStr(rstr_d);
}

const char *DpmError::GenerateBrief()
{
	if (extra.length()) {
		rstr_b = extra;
	} else if (my_serrno) {
		rstr_b = sstrerror(my_serrno);
	} else {
		rstr_b = "Unknown error";
	}
	return SafeCStr(rstr_b);
}


/******************************************************************************/
/*                D p m I d e n t i t y E x c e p t i o n                     */
/******************************************************************************/

DpmIdentityException::DpmIdentityException(int errnum, const char *detail)
	: errnum(errnum), detailmsg(detail)
{
	errmsg.assign(sstrerror(errnum),0);
}

const char *DpmIdentityException::what() const throw()
{
	return "DpmIdentityException";
}

DpmIdentityException::~DpmIdentityException() throw() {}

int DpmIdentityException::ErrNum() const { return errnum; }

const char *DpmIdentityException::ErrStr() const { return SafeCStr(errmsg); }

const char *DpmIdentityException::DetailStr() const
{
	return SafeCStr(detailmsg);
}

/******************************************************************************/
/*                            D p m I d e n t i t y                           */
/******************************************************************************/

void DpmIdentity::parse_secent(const XrdSecEntity *secEntity)
{
	m_name.erase();
	m_endors_raw.erase();

	if (!secEntity || !secEntity->name) {
		throw DpmIdentityException(EACCES,
			"Insufficient authentication data");
	}
	if (!strcmp(secEntity->prot,"sss")) {
		if (strcmp(secEntity->name,"nobody")) {
			try {
				m_name = DecodeString(secEntity->name);
			} catch(const char *e) {
				throw DpmIdentityException(EINVAL,
					"Error unencoding identity name "
					"from sss authentication protocol");
			}
		}
	} else if (!strcmp(secEntity->prot,"pwd")) {
		try {
			m_name = DecodeString(secEntity->name);
		} catch(const char *e) {
			throw DpmIdentityException(EINVAL,
				"Error unencoding identity name "
				"from pwd authentication protocol");
		}
	} else {
		m_name.assign(secEntity->name,0);
	}

	if (m_name.length() == 0) {
		throw DpmIdentityException(EACCES,
			"No identity provided by the authentication library");
	}

	if (!strcmp(secEntity->prot,"gsi")) {
		m_endors_raw.assign(secEntity->endorsements,0);
	} else if (!strcmp(secEntity->prot,"sss")) {
		if (secEntity->grps && strcmp(secEntity->grps,"nogroup")) {
			m_endors_raw.assign(secEntity->grps,0);
		}
	} else {
		m_endors_raw.assign(secEntity->grps,0);
	}
}

void DpmIdentity::check_validvo(DpmIdentityConfigOptions &config)
{
	if (m_name == "root") {
		return;
	}
	if (!config.validvo.size()) {
		return;
	}
	if (m_vorgs.size() == 0) {
		throw DpmIdentityException(EACCES,
			"User identity includes no vo to "
			"check against allowed list");
	}
	std::vector<XrdOucString>::const_iterator itr;
	bool ok=1;
	for(itr = m_vorgs.begin(); itr != m_vorgs.end(); ++itr) {
		if (std::find(config.validvo.begin(),
				config.validvo.end(), *itr) == 
					config.validvo.end()) {
			ok=0;
			break;
		}
	}
	if (!ok) {
		throw DpmIdentityException(EACCES,
			"User belongs to a vo which is not accepted");
	}
}

void DpmIdentity::parse_grps()
{
	XrdOucString vat;
	vs.clear();
	m_vorgs.clear();
	int from = 0;
	while ((from = m_endors_raw.tokenize(vat, from, ',')) != -1) {
		if (vat.length() <= 1)
			throw DpmIdentityException(EINVAL,
				"Group is too short");
		if (vat[0] != '/')
			throw DpmIdentityException(EINVAL,
				"Group does not start with /");
		XrdOucString vo;
		int isl = vat.find('/', 1);
		if (isl != STR_NPOS) {
			if (isl > 1) vo.assign(vat, 1, isl - 1);
		} else {
			vo.assign(vat, 1, vat.length() - 1);
		}
		if (!vo.length())
			throw DpmIdentityException(EINVAL,
				"Group includes no vo name");
		if (std::find(m_vorgs.begin(),m_vorgs.end(),vo)
				== m_vorgs.end()) {
			m_vorgs.push_back(vo);
		}
		isl = vat.find("/Role=NULL");
		if (isl != STR_NPOS)
			vat.erase(isl);
		isl = vat.find("/Capability=NULL");
		if (isl != STR_NPOS)
			vat.erase(isl);
		if (vat.length() > CA_MAXCSECNAMELEN)
			throw DpmIdentityException(ENAMETOOLONG,
				"Group is too long");
		vs.push_back(vat);
	}
}

DpmIdentity::DpmIdentity() :
	m_name("root"), m_grp_array(0), UsesSecEntForID(0) {}

DpmIdentity::DpmIdentity(XrdOucEnv *Env) : m_grp_array(0)
{
	UsesSecEntForID = 1;
	if (Env && Env->Get("dpm.dn"))
		UsesSecEntForID = 0;
	if (UsesSecEntForID) {
		const XrdSecEntity *secEntity = 0;
		if (Env)
			secEntity = Env->secEnv();
		parse_secent(secEntity);
	} else {
		try {
			m_name = DecodeString(Env->Get("dpm.dn"));
			m_endors_raw = DecodeString(Env->Get("dpm.voms"));
		} catch(const char *e) {
			throw DpmIdentityException(EINVAL,
				"Error unencoding identity strings "
				"from environment");
		}
	}
	parse_grps();
	if (m_name.length() == 0) {
		if (UsesSecEntForID) {
			throw DpmIdentityException(EACCES,
				"No identity provided by the authentication "
				"library");
		} else {
			throw DpmIdentityException(EACCES,
				"No identity passed in the environment");
		}
	}
}

DpmIdentity::DpmIdentity(XrdOucEnv *Env, DpmIdentityConfigOptions &config)
	: m_grp_array(0)
{
	UsesSecEntForID = 1;
	if (usesPresetID(Env))
		UsesSecEntForID = 0;
	if (UsesSecEntForID) {
		const XrdSecEntity *secEntity = 0;
		if (Env)
			secEntity = Env->secEnv();
		parse_secent(secEntity);
		parse_grps();
	} else {
		m_name = config.principal;
		if (m_name.length() == 0) {
			throw DpmIdentityException(EACCES,
				"No useable identity provided");
		}
		if (m_name.length() > CA_MAXCSECNAMELEN) {
			throw DpmIdentityException(ENAMETOOLONG,
				"Username too long");
		}
		if (m_name != "root") {
			std::vector<XrdOucString>::iterator itr;
			for(itr = config.fqans.begin();
					itr != config.fqans.end(); ++itr) {
				if (m_endors_raw.length())
					m_endors_raw += ",";
				m_endors_raw += *itr;
			}
			parse_grps();
		}
	}
	check_validvo(config);
}

/* static member function */
bool DpmIdentity::badPresetID(DpmIdentityConfigOptions &config,
		XrdOucString &err) {
	try {
		DpmIdentity id((XrdOucEnv*)0, config);
	} catch (DpmIdentityException &e) {
		err = e.DetailStr();
		return 1;
	}
	return 0;
}

DpmIdentity::~DpmIdentity() {
	DpmError dpmerr("DpmIdentity::~DpmIdentity");
	UnsetAuth();
}

void DpmIdentity::UnsetAuth()
{
	DpmError dpmerr("DpmIdentity::UnsetAuth");

	if (m_grp_array) {
		dpns_client_resetAuthorizationId();
		dpm_client_resetAuthorizationId();
		delete[] m_grp_array;
	}
	m_grp_array = 0;
	if (dpm_sf.has_thenv) {
		dpm_sf.th_setenv("CSEC_MECH",SafeCStr(dpm_sf.orig_mechstr),1);
	}
}

void DpmIdentity::SetAuth()
{
	DpmError dpmerr("DpmIdentity::SetAuth");

	int ngrps = 0;
        uid_t uid;
        gid_t gid0;

	if (dpm_sf.has_thenv) {
		dpm_sf.th_setenv("CSEC_MECH","ID",1);
	}

	if (dpns_client_resetAuthorizationId() < 0) {
		dpmerr.save("dpns_client_resetAuthorizationId");
		throw DpmIdentityException(dpmerr.GetSerrno(),
			dpmerr.GenerateDetail());
	}
	if (dpm_client_resetAuthorizationId() < 0) {
		dpmerr.save("dpm_client_resetAuthorizationId");
		throw DpmIdentityException(dpmerr.GetSerrno(),
			dpmerr.GenerateDetail());
	}

	if (m_grp_array) {
		delete[] m_grp_array;
		m_grp_array = 0;
	}

	if (strcmp(SafeCStr(m_name), "root")) {
		m_grp_array = new const char*[vs.size()+1];
		for (std::vector<XrdOucString>::const_iterator itr =
				vs.begin(); itr != vs.end(); ++itr) {
			m_grp_array[ngrps++] = SafeCStr(*itr);
		}
		m_grp_array[ngrps] = 0;
		// no need to find or store the actual virtual uid, gid(s)
		uid = gid0 = -1;

		if (dpns_client_setAuthorizationId(uid, gid0, "GSI",
				(char*)SafeCStr(m_name)) < 0) {
			dpmerr.save("dpns_client_setAuthorizationId");
			throw DpmIdentityException(dpmerr.GetSerrno(),
				dpmerr.GenerateDetail());
		}
		if (dpm_client_setAuthorizationId(uid, gid0, "GSI",
				(char*)SafeCStr(m_name)) < 0) {
			dpmerr.save("dpm_client_setAuthorizationId");
			throw DpmIdentityException(dpmerr.GetSerrno(),
				dpmerr.GenerateDetail());
		}
		if (ngrps) {
			if (dpns_client_setVOMS_data(
					(char*)SafeCStr(m_vorgs[0]),
					(char**)m_grp_array, ngrps) < 0) {
				dpmerr.save("dpns_client_setVOMS_data");
				throw DpmIdentityException(dpmerr.GetSerrno(),
					dpmerr.GenerateDetail());
			}
			if (dpm_client_setVOMS_data(
					(char*)SafeCStr(m_vorgs[0]),
					(char**)m_grp_array, ngrps) < 0) {
				dpmerr.save("dpm_client_setVOMS_data");
				throw DpmIdentityException(dpmerr.GetSerrno(),
					dpmerr.GenerateDetail());
			}
		}
	}
}

/******************************************************************************/
/*                        static member function                              */
/*                                                                            */
/* check if to use the preset identity given in the configuration             */
/* ==> dpm.principal <== and optional ==> dpm.fqan <== options.               */
/* If the preset is to be used a secondary authorization library must also    */
/* authorize an action.                                                       */
/******************************************************************************/
bool DpmIdentity::usesPresetID(XrdOucEnv *Env, const XrdSecEntity *Entity) {
	if (!Entity && Env)
		Entity = Env->secEnv();

	// if no authentication data, must use preset
	//
	if (!Entity)
		return 1;

	// if authentication type is [unix] then only consider the fixed
	// identity
	//
	if (!strcmp(Entity->prot, "unix"))
		return 1;

	// if it's the shared secret then check if a username is specified,
	// if not use the preset
	//
	if (!strcmp(Entity->prot, "sss")) {
		if (!Entity->name || !strcmp(Entity->name,"nobody")) {
			return 1;
		}
	}

	// if the environment has one of the alice token specifiers then use
	// the preset
	//
	if (Env && (Env->Get("signature") || Env->Get("authz")))
		return 1;

	// use the identity information in secEnt: the DpmIdentity constructor
	// has to be able to extract and parse the identity
	//
	return 0;
}

const XrdOucString &DpmIdentity::Dn() const { return m_name; }

const XrdOucString &DpmIdentity::Groups() const { return m_endors_raw; }

bool DpmIdentity::IsSecEntID() const { return UsesSecEntForID; }

/******************************************************************************/
/*                                functions                                   */
/******************************************************************************/

XrdOucString DecodeString(XrdOucString in) {
	XrdOucString tmp,out;
	int isl,from=0;

	while(1) {
		isl = in.find('%',from);
		if (isl == STR_NPOS)
			break;
		if (isl+3 > in.length())
			throw "DecodeString: not enough characters after "
				"percent";
		tmp.assign(in, isl+1, isl+2);
		unsigned int i;
		if (sscanf(SafeCStr(tmp), "%2x", &i) != 1 ||
				i < 0x20 || i > 0x7f)
			throw "DecodeString: unexpected character encoding";
		if (isl>0) {
			tmp.assign(in, from, isl-1);
			out.append(tmp);
		}
		out.append((char)i);
		from = isl + 3;
	}

	tmp.assign(in,from);
	out.append(tmp);

	return out;
}


// Encode an array of bytes to base64

char *Tobase64(const unsigned char *input, int length) {
	BIO *bmem, *b64;
	BUF_MEM *bptr;

	b64 = BIO_new(BIO_f_base64());
	bmem = BIO_new(BIO_s_mem());
	b64 = BIO_push(b64, bmem);
	BIO_write(b64, input, length);

	if (BIO_flush(b64) <= 0) {
		BIO_free_all(b64);
		return NULL;
	}

	BIO_get_mem_ptr(b64, &bptr);

	char *buff = (char *) malloc(bptr->length);
	memcpy(buff, bptr->data, bptr->length - 1);
	buff[bptr->length - 1] = 0;

	BIO_free_all(b64);

	return buff;
}

// Calculates the url hash, from the given information
// The hash is calculated from:
// - xrd_fn, the filename in the xroot request
// - surl of the file (needed elsewhere to close it)
// - dpmdhost
// - sfn
// - dpm request token
// - flags, set of bit mapped flags
// - DN of the client
// - voms string
// - creation time of the url
// - validity time before and after creation time
// - nonce string
// - key
// - key length
//

char *calcHash(
	const char *xrd_fn,
	const char *surl,
	const char *dpmdhost,
	const char *sfn,
	const char *dpmtk,
	unsigned int flags,
	const char *dn,
	const char *vomsnfo,
	time_t tim,
	int tim_grace,
	const char *nonce,
	unsigned const char *key,
	size_t keylen) {

	int           i;
	HMAC_CTX      ctx;
	unsigned int  len;
	unsigned char mdbuf[EVP_MAX_MD_SIZE];
	char buf[64];
	struct tm tms;
	
	if (!xrd_fn || !surl || !dpmdhost ||
			!sfn ||!dpmtk || !dn || !vomsnfo || !nonce) {
		return 0;
	}
   
	HMAC_CTX_init(&ctx);
	HMAC_Init_ex(&ctx, key, keylen, EVP_sha256(), 0);
	
	HMAC_Update(&ctx, (const unsigned char *)xrd_fn, strlen(xrd_fn)+1);
	HMAC_Update(&ctx, (const unsigned char *)surl, strlen(surl)+1);
	HMAC_Update(&ctx, (const unsigned char *)dpmdhost, strlen(dpmdhost)+1);
	HMAC_Update(&ctx, (const unsigned char *)sfn, strlen(sfn)+1);	
	HMAC_Update(&ctx, (const unsigned char *)dpmtk, strlen(dpmtk)+1);
	
	sprintf(buf, "%u", flags);
	HMAC_Update(&ctx, (unsigned char *)buf, strlen(buf)+1);
	
	HMAC_Update(&ctx, (const unsigned char *)dn, strlen(dn)+1);
	HMAC_Update(&ctx, (const unsigned char *)vomsnfo, strlen(vomsnfo)+1);

	localtime_r(&tim, &tms);
	strftime(buf, sizeof(buf), "%s", &tms);	
	sprintf(&buf[strlen(buf)],",%d", tim_grace);
	HMAC_Update(&ctx, (unsigned char *)buf, strlen(buf)+1);
	HMAC_Update(&ctx, (unsigned char *)nonce, strlen(nonce)+1);
	
	HMAC_Final(&ctx, mdbuf, &len);
	HMAC_CTX_cleanup(&ctx);
	
	return Tobase64(mdbuf, len/2);
}

int compareHash(
	const char *h1,
	const char *h2)
{
	if (!h1 || !h2)
		return 1;

	size_t l1 = strlen(h1), l2 = strlen(h2);
	if (l1 != l2)
		return 1;

	char result = 0;
	for(size_t i = 0; i < l1; ++i) {
		result |= h1[i] ^ h2[i];
	}
	return result != 0;
}

// The syntax of the sfn is a bit funky for an xrootd url:
//  host:/path
// Here we strip the first token, if a semicolon is there
const char *getfnamestartfromsfn(const char *sfn) {

        const char *pos = strchr(sfn, ':');
        const char *pos2 = strchr(sfn, '/');
        if (pos && (!pos2 || pos2>pos)) return pos+1;
        return sfn;
}

void InitLocalHostNameList(std::vector<XrdOucString> &names) {
	char localhost[CA_MAXHOSTNAMELEN+1];

	names.clear();
	localhost[0] = '\0';
	if (!gethostname(localhost, sizeof(localhost))) {
		if (strchr(localhost, '.') == 0) {
			char localdomain[CA_MAXHOSTNAMELEN+1];
		        if (!Cdomainname(localdomain, sizeof(localdomain)) &&
				(strlen(localhost) + strlen(localdomain)
						< CA_MAXHOSTNAMELEN)) {
		                strcat (localhost, ".");
		                strcat (localhost, localdomain);
			} else {
				localhost[0] = '\0';
			}
	        }
	}
	if (strlen(localhost))
		names.push_back(localhost);

	const char *p = getenv("DPMXRD_ALTERNATE_HOSTNAMES");
	if (p) {
		char *p3,*p2 = strdup(p);
		char *curr = p2;
		while((p3=strsep(&curr, " ,\t"))) {
			names.push_back(p3);
		}
		free(p2);
	}
}

const char *LoadKeyFromFile(unsigned char **dat, size_t *dsize) {
	int fd;
	struct stat sbuf;
	const static char fn[] = "/etc/xrootd/dpmxrd-sharedkey.dat";
	class lkh_c {
	public:
		lkh_c(int fd) : fd(fd) { };
		~lkh_c() { close(fd); }
	private:
		int fd;
	};

	if (!dat || !dsize) {
		return "Invalid argument";
	}

	*dat = 0;
	*dsize = 0;

	fd = open(fn, O_RDONLY);

	if (fd < 0) {
		return fn;
	}
	lkh_c hlp(fd);

	if (fstat(fd, &sbuf)<0) {
		return "Could not stat";
	}

	if (!S_ISREG(sbuf.st_mode)) {
		return "Not a regular file";
	}

	if (sbuf.st_mode & (S_IROTH | S_IWOTH | S_IXOTH)) {
		return "'Other' user has some permissions set";
	}

	if (sbuf.st_size == 0) {
		return "Empty";
	}

	if (sbuf.st_size > 64) {
		return "Too long";
	}

	size_t remain = sbuf.st_size;
	size_t nread;
	ssize_t nret;

	*dat = (unsigned char *)malloc(remain);
	nread = 0;
	while(remain) {
		nret = read(fd, (*dat + nread), remain);
		if (nret < 0) {
			free(*dat);
			*dat = 0;
			return "Could not read the shared key";
		} else if (nret == 0) {
			free(*dat);
			*dat = 0;
			return "Unexpected end of file";
		} else {
			nread += nret;
			remain -= nret;
		}
	}

	*dsize = nread;
	return 0;
}

XrdOucString CanonicalisePath(const char *s, int trailing_slash)
{
        XrdOucString out;
        out.assign(s,0);

        while(out.find("//") != STR_NPOS)   
                out.replace("//","/");
 	if (!out.beginswith('/'))
                out = "/" + out;
        if (trailing_slash && !out.endswith('/'))
                out += "/";
        else if (!trailing_slash && out.length()>1 && out.endswith('/'))
                out.erasefromend(1);

        return out;
}

XrdOucString TranslatePath(
                DpmRedirConfigOptions &config,
         	const char *in)
{
	char Apath[MAXPATHLEN*2+1];

	if (config.theN2N) {
		int ret;
		Apath[MAXPATHLEN*2] = '\0';
		XrdOucString envstr;
		const char *p;

		if (dpm_sf.has_thenv) {
			if ((p = dpm_sf.th_getenv("CSEC_MECH"))) {
				envstr.assign(p,0);
			}
			dpm_sf.th_setenv("CSEC_MECH",
				SafeCStr(dpm_sf.orig_mechstr),1);
		}
		
		if (!(ret=config.theN2N->lfn2pfn(in, Apath, sizeof(Apath)))) {
			in = Apath;
		}

		if (dpm_sf.has_thenv) {
			dpm_sf.th_setenv("CSEC_MECH",SafeCStr(envstr),1);
		}

		if (ret || Apath[MAXPATHLEN*2]) {
			if (ret) {
				throw -ret;
			}
			throw ENAMETOOLONG;
		}
		XrdOucString check = CanonicalisePath(in,1);
		bool ok=0;
		std::vector<XrdOucString>::const_iterator itr;
		for(itr=config.N2NCheckPrefixes.begin();
				itr!=config.N2NCheckPrefixes.end();++itr) {
			if (check.find(*itr) == 0) {
				ok=1;
				break;
			}
		}
		if (!ok) {
			throw EACCES;
		}
	}

 	bool endsWithSlash = (strlen(in) && in[strlen(in)-1] == '/');
        XrdOucString out = CanonicalisePath(in, 1);

	if (config.theN2N) {
	        if (!endsWithSlash)
	                out.erasefromend(1);
                return(out);
	}

        std::vector<std::pair<XrdOucString, XrdOucString > >::const_iterator
		itr;

        for(itr=config.pathPrefixes.begin(); itr != config.pathPrefixes.end();
			++itr) {
                const XrdOucString &pfx = itr->first;
                const XrdOucString &repl = itr->second;
                if (out.find(pfx) == 0) {
                        out.erasefromstart(pfx.length());
                        out = repl + out;
                        if (!endsWithSlash)
                                out.erasefromend(1);
                        return out;
                }
        }

        if (config.defaultPrefix.length() == 0 ||
                        out.find(config.defaultPrefix) == 0) {
	        if (!endsWithSlash)
	                out.erasefromend(1);
                return(out);
        }

        out.erasefromstart(1);
        out = config.defaultPrefix + out;
        if (!endsWithSlash)
                out.erasefromend(1);

        return out;
}
