/* 
 * File:   XrdDPMDiskAcc.cc
 * Author: Fabrizio Furano (CERN IT/GT)
 * 
 * An Oss plugin that interacts with a DPM
 *
 * Created on Jan 13, 2012, 10:25 AM
 */

#include "XrdDPM/XrdDPMDiskAcc.hh"
#include "XrdOss/XrdOssTrace.hh"
#include "XrdOuc/XrdOucEnv.hh"
#include "XrdSec/XrdSecEntity.hh"


DFThreadSpecific dpm_sf;


/******************************************************************************/
/*                  E r r o r   R o u t i n g   O b j e c t                   */
/******************************************************************************/
   

  
XrdSysError OssEroute(0, "DPMDiskAcc_");
   
XrdOucTrace OssTrace(&OssEroute);




XrdAccPrivs XrdDPMDiskAcc::Access(const XrdSecEntity    *Entity,
	const char            *path,
	const Access_Operation oper,
	XrdOucEnv       *env) {
	
	const char *tident="tident";
	EPNAME("XrdDPMDiskAcc::Access");
	
	const char *dpmtk = 0;
	const char *dpmnonce = 0;
	char *isput_s = 0;
	unsigned int flags = 0;
	const char *dhost = 0;
	const char *tim_s = 0;
	time_t tim = 0;
	int reqgracetime = 0;
	int effgracetime = 0;
	const char *p;
	XrdOucString sfn, surl, dn, vomsnfo;
	struct tm tms;

	if (!env) {
                TRACE(ALL,"No environment parameters passed.");
                return XrdAccPriv_None;
        }

	if (oper != AOP_Create && oper != AOP_Update && oper != AOP_Read) {
		DEBUG("Opperation " << oper << " not permitted.");
		return XrdAccPriv_None;
	}

	// Then, get all the other parameters to verify

	try {
		sfn = DecodeString(env->Get("dpm.sfn"));
		surl = DecodeString(env->Get("dpm.surl"));
	} catch (const char *e) {
		DEBUG("Error unencoding cgi argument: " << e);
		return XrdAccPriv_None;
	}

	if ((isput_s = env->Get("dpm.put")) && atoi(isput_s))
		flags |= XRDDPM_IS_PUT;
	
	dpmtk = env->Get("dpm.tk");
	dpmnonce = env->Get("dpm.nonce");

	dhost = env->Get("dpm.dhost");
	if (!dhost) {   
                DEBUG("Token is missing target disk server host name.");
                return XrdAccPriv_None;
        }

	// Get the req time, do some checks
	tim_s = env->Get("dpm.time");
	if (!tim_s || !(tim_s = strptime(tim_s, "%s", &tms))
			|| (*tim_s && *tim_s!=','))  {
		DEBUG("Invalid request time.");
		return XrdAccPriv_None;
	}
	tim = mktime(&tms);
	if (*tim_s)
		reqgracetime = atoi(++tim_s);
	if (reqgracetime<0) {
		DEBUG("Negative validity time in request.");
		return XrdAccPriv_None;
	}
	effgracetime = reqgracetime;
	if (maxgracetime) {
		if (!effgracetime || effgracetime > maxgracetime) {
			effgracetime = maxgracetime;
		}
	}
	if (!effgracetime) {
		XrdOucString err =
			"No maximum validity period configured or supplied in request.";
		OssEroute.Emsg("Access", SafeCStr(err));
		return XrdAccPriv_None;
	}
	if (tim < time(0)-effgracetime) {
		TRACE(ALL, "Request time expired.");
		return XrdAccPriv_None;
	}
	if (tim > time(0)+effgracetime) {
		XrdOucString err = "Request time is in the future.";
		OssEroute.Emsg("Access", SafeCStr(err));
		return XrdAccPriv_None;
	}

	try {
		DpmIdentity dpmid(env);
		dn = dpmid.Dn();
		vomsnfo = dpmid.Groups();
		if (!dpmid.IsSecEntID()) {
			flags |= XRDDPM_ID_URL;
		}
	} catch(DpmIdentityException &e) {
		DEBUG("Could not get identity: " << e.DetailStr());
		return XrdAccPriv_None;
	}

	char *hash2 = env->Get("dpm.hv1");
	if (!hash2) {
		DEBUG("Hash v1 not found.");
		return XrdAccPriv_None;
	}
	
	// Now use the shared secret key to calculate the hash from the cgi info
	char *hash = 0;
	if (key) {
		
		hash = calcHash(path, SafeCStr(surl), dhost, SafeCStr(sfn), dpmtk ? dpmtk : "",
			flags, SafeCStr(dn), SafeCStr(vomsnfo), tim, reqgracetime,
			dpmnonce ? dpmnonce : "", key, keySize);
		if (!hash) {
			XrdOucString err = "Failed to calculate hash.";
			OssEroute.Emsg("Access", SafeCStr(err));
			return XrdAccPriv_None;
		}

	}
	else {
		XrdOucString err = "No shared secret --> no hash.";
		OssEroute.Emsg("Access", SafeCStr(err));
		return XrdAccPriv_None;
	}
	
	if (compareHash(hash, hash2)) {
		XrdOucString err = "Invalid request signature.";
		free(hash);
		if (Entity && Entity->host && *Entity->host) {
			err += " Client host " + XrdOucString(Entity->host);
		}
		OssEroute.Emsg("Access", SafeCStr(err));
		return XrdAccPriv_None;
	}
	free(hash);

	std::vector<XrdOucString>::const_iterator itr;
	bool dfound=0;
	for(itr = LocalHostNames.begin();
			itr != LocalHostNames.end(); ++itr) {
		if (!strcasecmp(SafeCStr(*itr),dhost)) {
			dfound=1;
			break;
		}
	}
	if (!dfound) {
		XrdOucString err = "Disk server hostname "
			+ XrdOucString(dhost)
			+ " not matched to this host.";
		OssEroute.Emsg("Access", SafeCStr(err));
		return XrdAccPriv_None;
	}

	return (flags & XRDDPM_IS_PUT) ? XrdAccPriv_Create : XrdAccPriv_Read;
}





int XrdDPMDiskAcc::Test(const XrdAccPrivs priv,const Access_Operation oper)
{

// Warning! This table must be in 1-to-1 correspondence with Access_Operation
//
   static XrdAccPrivs need[] = {XrdAccPriv_None,                 // 0
                                XrdAccPriv_Chmod,                // 1
                                XrdAccPriv_Chown,                // 2
                                XrdAccPriv_Create,               // 3
                                XrdAccPriv_Delete,               // 4
                                XrdAccPriv_Insert,               // 5
                                XrdAccPriv_Lock,                 // 6
                                XrdAccPriv_Mkdir,                // 7
                                XrdAccPriv_Read,                 // 8
                                XrdAccPriv_Readdir,              // 9
                                XrdAccPriv_Rename,               // 10
                                XrdAccPriv_Lookup,               // 11
                                XrdAccPriv_Update                // 12
                               };
   if (oper < 0 || oper > AOP_LastOp) return 0;
   return (int)(need[oper] & priv) == need[oper];
}




extern "C" XrdAccAuthorize *XrdAccAuthorizeObject(XrdSysLogger *lp,
                                                     const char   *cfn,
                                                     const char   *parm) {
	
	const char *tident="tident";
	EPNAME("XrdAccAuthorizeObject");

	XrdSysError Say(0,"DPMDiskAccLoader_");
	Say.logger(lp);

	XrdDPMDiskAcc *myacc = new XrdDPMDiskAcc();
	size_t thekeylen = 0;

	if (myacc && parm) {
		// Parse the parameters, extracting
		// - shared secret key
		// - grace period
		char buf[256];
		const char *p = parm;
		for ( ; *p && (*p == ' '); p++);
		const char *p2 = p;
		int tkcnt = 0;

		while(p2) {
			p2 = strchr(p, ' ');
			memset(buf, 0, sizeof(buf));

			if (p2) strncpy(buf, p, p2-p);
			else strcpy(buf, p);

			switch(tkcnt) {
				case 0:
					Say.Emsg("NewObject","setting maxgracetime:", buf);
					myacc->setMaxGracetime(atoi(buf));
					break;
				case 1:
					Say.Emsg("NewObject","setting new key");
					thekeylen = strlen(buf);
					myacc->setKey((unsigned char *)buf, thekeylen);
					break;
			}
			tkcnt++;

			if (p2) {
				p = p2+1;
				for ( ; *p && (*p == ' '); p++);
			}

		}
	}

	if (!thekeylen) {
		unsigned char *dat;
		const char *msg;
		if ((msg = LoadKeyFromFile(&dat, &thekeylen))) {
			Say.Emsg("NewObject","Error while reading key from file:", msg);
			delete myacc;
			myacc = 0;
		} else {
			myacc->setKey(dat, thekeylen);
			free(dat);
		}
	}

	if (myacc)
		return myacc;

	Say.Emsg("NewObject","cannot create DPMDisk Acc instance");
	return 0;
}
