/*                         
 * Copyright (C) 2012 by CERN/IT/GT/DMS. All rights reserved.
 *
 * Author: Fabrizio Furano, David Smith <David.Smith@cern.ch>
*/

#include "XrdDPM/XrdDPMFinder.hh"
#include "XrdDPM/XrdDPMRedirTrace.hh"
#include "XrdOuc/XrdOucStream.hh"
#include "XrdOuc/XrdOuca2x.hh"
#include "XrdSec/XrdSecEntity.hh"
#include "XrdSfs/XrdSfsInterface.hh"
#include "XrdSys/XrdSysPlugin.hh"
#include "XrdSys/XrdSysDNS.hh"
#include "Xrd/XrdLink.hh"
#include "Xrd/XrdLinkMatch.hh"
#include "XrdSut/XrdSutRndm.hh"

#include <sys/time.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <string.h>

#include "serrno.h"

#define DPMRETRY(x) ((x)==SECOMERR || (x)==SEINTERNAL)
#define XPROTO "xroot"
#define MAX_QLIFE 1800
#define QABORT_LIMIT 900
#define MAX_BACKOFF 180
#define META_RETRYTIME 30
#define MIN_KEYLEN 32

// globals
namespace DpmFinder {
	XrdSysError	Say(0,"dpmfinder_");
	XrdOucTrace	Trace(&Say);
}

using namespace DpmFinder;

DFThreadSpecific dpm_sf;

class XrdOss;
extern XrdOucTrace OfsTrace;

// static functions, objects local to this compilation unit
//

static XrdOucString SockaddrToIP(const sockaddr *sa) {
	char buf[INET6_ADDRSTRLEN+2];
	XrdOucString ip;

	if (!sa) return ip;
	if (XrdSysDNS::IPFormat(sa, buf, sizeof(buf), 0))
		ip.assign(buf,0);

	return ip;
}

static XrdOucString EncodeString(const XrdOucString &in) {
	XrdOucString out(in);
	static const char esc[] = { '&', '?', ' ', '\0' };

	out.replace("%","%25");
	for(const char *p=esc;*p;++p) {
		char cs[2],ct[4];
		sprintf(cs,"%c",*p);    
		sprintf(ct,"%%%02x",*p);
		out.replace(cs,ct);
	}

	return out;
}

static XrdOucString PrintStatus(int status) {
	XrdOucString out;
	int idx = (status&0xf000)>>12;
	static const char *names[] = { "DPM_SUCCESS", "DPM_QUEUED",
		"DPM_ACTVIE", "DPM_READY", "DPM_RUNNING", "DPM_DONE",
		"DPM_FAILED", "DPM_ABORTED", "DPM_EXPIRED", "DPM_RELEASED",
		"DPM_TO_BE_MIGRATED", "DPM_TO_BE_RECALLED" };
	if (idx < 12) {
		out = names[idx];
		if (status&0xfff) {
			out += "|";
			out += (status&0xfff);
		}
	} else {
		out = status;
	}
	return out;
}

static int getMode(const char *theMode, mode_t &Mode)
{
	char *eP;

	// Convert the mode argument
	//
	if (!(Mode = strtol(theMode, &eP, 8)) || *eP || (Mode >> 9)) return 0;
	return 1;
}

static bool checkDirExists(const XrdOucString &dname)
{
	EPNAME("checkDirExists");
	std::auto_ptr<DpmIdentity> identP;
	try {
		identP.reset(new DpmIdentity());
		identP->SetAuth();
	} catch(DpmIdentityException &e) {
		return 0;
	}

	DpmError dpmerr(epname);
	struct dpns_filestat statbuf;
	XrdOucString dn = CanonicalisePath(SafeCStr(dname),0);
	memset(&statbuf,'\0',sizeof(statbuf));
	if (!dpns_stat(SafeCStr(dn),&statbuf)) {
		if (S_ISDIR(statbuf.filemode))
			return 1;
	}
	return 0;
}

static XrdOucString mk_opaque(
		const char *xrd_fn,
		const DpmFileRequest &req,
		const DpmIdentity &ident,
		const std::vector<unsigned char> &key,
		int grace) {
	XrdOucString s,nonce,surl,sfn,rtoken;
	time_t t0 = time(0);
	char tbuf[64];
	struct tm tms;

	s = "&dpm.time=";
	localtime_r(&t0, &tms);
        strftime(tbuf, sizeof(tbuf), "%s", &tms);
	s += tbuf;
	if (grace) {
		sprintf(tbuf,",%d",grace);
		s += tbuf;
	}

	if (!ident.IsSecEntID()) {
		s += "&dpm.dn=" + EncodeString(ident.Dn());
		if (ident.Groups().length())
			s += "&dpm.voms=" + EncodeString(ident.Groups());
	}

	s += "&dpm.dhost=" + req.Host();

	if (req.ROpts().isPut && req.Sfn().length() &&
			strcmp(SafeCStr(req.Path()),xrd_fn)) {
		surl = req.Path();
		s += "&dpm.surl=" + EncodeString(surl);
	}

	if (req.Sfn().length()) {
		sfn = req.Sfn();
		s += "&dpm.sfn=" + EncodeString(
			getfnamestartfromsfn(SafeCStr(sfn)));
	}

	if (req.ROpts().isPut && req.RToken().length()) {
		rtoken = req.RToken();
		s += "&dpm.tk=" + rtoken;
	} else {
		if (XrdSutRndm::GetString(3,8,nonce))
			throw "Could not make a random nonce";
		s += "&dpm.nonce=" + nonce;
	}

	if (req.ROpts().isPut)
		s+= "&dpm.put=1";

	s += "&dpm.hv1=";

	char *hash = calcHash(
				xrd_fn,
				SafeCStr(surl),
				SafeCStr(req.Host()),
				getfnamestartfromsfn(SafeCStr(sfn)),
				SafeCStr(rtoken),
				(req.ROpts().isPut ? XRDDPM_IS_PUT : 0) |
					(ident.IsSecEntID() ? 0 : XRDDPM_ID_URL),
				SafeCStr(ident.Dn()),
				SafeCStr(ident.Groups()),
				t0,
				grace,
				SafeCStr(nonce),
				&key[0],
				key.size() );

	if (!hash)
		throw "Could not calculate redirection hash";

	s += hash;
	free(hash);

	return s;
}

static int mkp(const char *path, mode_t mode)
{
	EPNAME("mkp");
	char *next_path, this_path[CA_MAXPATHLEN+1];
	struct dpns_filestatg dpnsbuf;
	DpmError dpmerr(epname);
	int retc;

	DEBUG("Makepath " << path << " mode=" << std::oct << mode << std::dec);
          
	if (strlen(path)>CA_MAXPATHLEN)
		return ENAMETOOLONG;
	strcpy(this_path, path);
       
	// Trim off the trailing slash so that we make everything but the last
	// component
	//
	if (!(retc = strlen(this_path))) {
		return ENOENT;
	}
	while(retc && this_path[retc-1] == '/')
		{retc--; this_path[retc] = '\0';}
     
	// Typically, the path exists. So, do a quick check before launching
	// into it
	//
	if (!(next_path = rindex(this_path, (int)'/'))
	  ||  next_path == this_path) {
	return 0;
	}
	*next_path = '\0';
	if (!dpns_statg(this_path, 0, &dpnsbuf)) {
		return 0;
	}
	*next_path = '/';
   
	// Start creating directories starting with the root
	//
	next_path = this_path;
	while((next_path = index(next_path+1, int('/'))))
	{
		if (next_path[1] == '/') continue;
		*next_path = '\0';
		if (dpns_statg(this_path, 0, &dpnsbuf)<0) {
			if (serrno == ENOENT) {
				if ( dpns_mkdir(this_path, mode)<0 &&
						serrno != EEXIST ) {
					return serrno;
				}
			} else {
				return serrno;
			}
		}
		*next_path = '/';
        }
	// All done
	//
	return -1;
}

/******************************************************************************/
/*               D p m F i l e R e q u e s t E x c e p t i o n                */
/******************************************************************************/

DpmFileRequestException::DpmFileRequestException(int status, const char *brief,
	const char *detail): status(status), detailmsg(detail)
{
	if (brief && *brief) {
		errmsg.assign(brief,0);
	} else if (status&0xfff) {
		errmsg.assign(sstrerror(status&0xfff),0);
	} else {
		errmsg = "Failure during a dpm file request";
	}
}

DpmFileRequestException::~DpmFileRequestException() throw() {}

const char *DpmFileRequestException::what() const throw()
	{ return "DpmFileRequestException"; }

int DpmFileRequestException::DpmStatus() const { return status; }

int DpmFileRequestException::ErrCode() const { return (status&0xfff); }

const char *DpmFileRequestException::ErrStr() const { return SafeCStr(errmsg); }

const char *DpmFileRequestException::DetailStr() const
{
	if (detailmsg.length()) {
		return SafeCStr(detailmsg);
	} else {
		return SafeCStr(errmsg);
	}
}

/******************************************************************************/
/*                       D p m F i l e R e q u e s t                          */
/******************************************************************************/

DpmFileRequest::~DpmFileRequest() {}

int DpmFileRequest::LastStatus() const { return ReqStatus; }

void DpmFileRequest::UpdateStatus() {
	EPNAME("DpmFileRequest::UpdateStatus");
	bool isCreate = flags & (SFS_O_CREAT | SFS_O_TRUNC);
	bool userWantsOverwrite = (flags & SFS_O_TRUNC);

	if (ReqStatus == DPM_READY)
		return;

	if (MkpathState == 1) {
		int ret = mkp(SafeCStr(spath), 0775);
		if (ret>0) {
			if (DPMRETRY(ret))
				return;
			throw DpmFileRequestException((ReqStatus&0xf000)|ret,
				"Error making directories for file path");
		}
		MkpathState = 2;
	}
	if (ReqOpts.isPut) {

		if (isFirst && MkpathState==0 &&
				isAccessiblePutOngoing()) {
			DEBUG("Reusing sfn of existing put request");
			return;
		}

		if (!isCreate) {
			throw DpmFileRequestException(
				(ReqStatus&0xf000)|ENOTSUP,
				"Open for update not supported");
		}

		try {
			if (isFirst)
				do_first_put();
			else
				do_putstatus();

		} catch (DpmFileRequestException &e) {
			if (e.ErrCode() == ENOENT) {
				if (ReqOpts.isMkpath && isCreate &&
						MkpathState==0) {
					init();
					MkpathState = 1;
					UpdateStatus();
					return;
				}
			}
			if (e.ErrCode() == EEXIST ||
					(e.DpmStatus()&0xf000) == DPM_ABORTED) {
				if (isAccessiblePutOngoing()) {
					DEBUG("Reusing sfn of existing "
						"put request");
					return;
				}
			}
			if (e.ErrCode() == EEXIST &&
					!withOverwrite && userWantsOverwrite) {
				init();
				MkpathState = 2;
				withOverwrite = 1;
				UpdateStatus();
				return;
			}
			throw;
		}
	} else {
		if (isFirst)
			do_first_get();
		else
			do_getstatus();
	}
}

void DpmFileRequest::Abort() {
	EPNAME("DpmFileRequest::Abort");
	int nbreplies = 0;
	struct dpm_filestatus *filestatuses = 0;
	int ret;
	DpmError dpmerr(epname);
	char *surl = strdup(SafeCStr(spath));

	ret = dpm_abortfiles(r_token, 1, &surl, &nbreplies, &filestatuses);
	dpmerr.save("dpm_abortfiles");
	DEBUG("dpm_abortfiles(r_token=" << r_token << ",surl=" <<
		surl << ")=" << ret << ", serrno=" << dpmerr.GetSerrno());
	free(surl);
	dpm_free_filest(nbreplies, filestatuses);
	if (ret<0) {
		throw DpmFileRequestException(DPM_FAILED|(dpmerr.GetSerrno()),
			dpmerr.GenerateBrief(),dpmerr.GenerateDetail());
	}
}

int DpmFileRequest::BackoffWaitTime() {
	if (waittime) {
		waittime *= 2;
		if (waittime > MAX_BACKOFF) {
			waittime = MAX_BACKOFF;
		}
	} else {
		waittime = 1;
	}
	return waittime;
}

const XrdOucString &DpmFileRequest::Host() const { return r_host; }

const XrdOucString &DpmFileRequest::Sfn() const { return r_sfn; }

const XrdOucString &DpmFileRequest::Path() const { return spath; }

const XrdOucString DpmFileRequest::RToken() const
	{ return XrdOucString(r_token); }

const DpmFileRequestOptions &DpmFileRequest::ROpts() const { return ReqOpts; }

DpmFileRequest::DpmFileRequest(const XrdOucString &path, int flags,
	DpmFileRequestOptions &ReqOpts) : waittime(0), spath(path),
	flags(flags), ReqOpts(ReqOpts) { init(); }

void DpmFileRequest::init() {
	MkpathState = 0;
	isFirst = 1;
	withOverwrite = 0;
	r_token[0] = '\0';
	r_host.erase();
	r_sfn.erase();
	ReqStatus = 0;
}

DpmFileRequest::put_helper::put_helper(struct dpm_putfilereq *req,
	struct dpm_putfilestatus **statuses, int *reps)
	: req(req), statuses(statuses), reps(reps)
{
	memset(req,0,sizeof(struct dpm_putfilereq));
	*statuses = 0; *reps = 0;
}
DpmFileRequest::put_helper::~put_helper() {
	free(req->to_surl);
	dpm_free_pfilest(*reps, *statuses);
}
		
DpmFileRequest::get_helper::get_helper(struct dpm_getfilereq *req,
	struct dpm_getfilestatus **statuses, int *reps)
	: req(req), statuses(statuses), reps(reps)
{
	memset(req,0,sizeof(struct dpm_getfilereq));
	*statuses = 0; *reps = 0;
}

DpmFileRequest::get_helper::~get_helper() {
	free(req->from_surl);
	dpm_free_gfilest(*reps, *statuses);
}

void DpmFileRequest::do_first_put() {
	EPNAME("DpmFileRequest::do_first_put");
	struct dpm_putfilestatus *filestatuses;
	struct dpm_putfilereq reqfiles;
	char *pp,protocols[32],u_token[256];
	time_t retrytime = 0;
	int nbreplies;
	DpmError dpmerr(epname);

	put_helper h(&reqfiles, &filestatuses, &nbreplies);
	reqfiles.to_surl = strdup(SafeCStr(spath));
	reqfiles.lifetime = ReqOpts.lifetime;
 	reqfiles.f_type = ReqOpts.ftype;
	reqfiles.requested_size = ReqOpts.reqsize;

	strcpy(reqfiles.s_token, ReqOpts.stoken);
	// nb protocols array is fixed size
	strcpy(protocols, XPROTO);
	// nb u_token is fixed size
	strcpy(u_token,"DPM xrootd-Finder put request"); 

	int overwrite = withOverwrite ? 1 : 0;

	DEBUG("calling with to_surl='" << reqfiles.to_surl <<
		"', lifetime=" << reqfiles.lifetime << ", f_type='" <<
		((reqfiles.f_type) ? XrdOucString(reqfiles.f_type) :
			XrdOucString("") ) <<
		"', requested_size=" << reqfiles.requested_size <<
		", s_token='" << reqfiles.s_token <<
		"' overwrite=" << overwrite);

	if (overwrite) {
		TRACE(truncate, "Overwrite enabled on put request for "
			<< reqfiles.to_surl);
	}

	ReqStatus = 0;
	pp = protocols;
	if (dpm_put(1, &reqfiles, 1, &pp, u_token, overwrite,
			retrytime, r_token, &nbreplies, &filestatuses)<0) {
		dpmerr.save("dpm_put");
		ReqStatus = DPM_FAILED | serrno;
		TRACE(write,"dpm_put failed with to_surl='" << reqfiles.to_surl
			<< "' => serrno=" << serrno
			<< " detail=" << dpmerr.GenerateDetail());
		if (filestatuses && filestatuses[0].errstring)
			dpmerr.assign(filestatuses[0].errstring);
	} else if (nbreplies) {
		ReqStatus = filestatuses[0].status;
		TRACE(write,"put result for to_surl='" << reqfiles.to_surl <<
			"' => r_token=" << r_token <<
			", status=" << PrintStatus(ReqStatus));
		if (ReqStatus == DPM_READY)
			parse_turl(filestatuses[0].turl);
	} else {
		ReqStatus = DPM_FAILED | EINVAL;
		TRACE(write,"No file statuses returned");
	}

	if (ReqStatus != DPM_QUEUED && ReqStatus != DPM_READY &&
			ReqStatus != DPM_ACTIVE) {
		if ((ReqStatus&0xf000)==DPM_FAILED &&
				DPMRETRY(ReqStatus&0xfff)) {
			// perhaps next time the request will work
			TRACE(write,"Can be retried");
			return;
		}
		if (dpmerr.GetSerrno()) {
			throw DpmFileRequestException(ReqStatus,
				dpmerr.GenerateBrief(),dpmerr.GenerateDetail());
		} else {
			throw DpmFileRequestException(ReqStatus);
		}
	}
	isFirst = 0;
}

void DpmFileRequest::do_putstatus() {
	EPNAME("DpmFileRequest::do_putstatus");
	struct dpm_putfilestatus *filestatuses;
	struct dpm_putfilereq reqfiles;
	int nbreplies;
	DpmError dpmerr(epname);

	ReqStatus = 0;
	put_helper h(&reqfiles, &filestatuses, &nbreplies);

	reqfiles.to_surl = strdup(SafeCStr(spath));

	DEBUG("calling with r_token=" << r_token <<
		", to_surl='" << reqfiles.to_surl << "'");

	if (dpm_getstatus_putreq(r_token, 1, &reqfiles.to_surl, &nbreplies,
			&filestatuses)<0) {
		dpmerr.save("dpm_getstatus_putreq");
		ReqStatus = DPM_FAILED | serrno;
		TRACE(write,"dpm_getstatus_putreq failed with r_token="
			<< r_token << ", to_surl='" << reqfiles.to_surl
			<< "' => serrno=" << serrno
			<< " detail=" << dpmerr.GenerateDetail());
		if (filestatuses && filestatuses[0].errstring)
			dpmerr.assign(filestatuses[0].errstring);
	} else if (nbreplies) {
		ReqStatus = filestatuses[0].status;
		TRACE(write,"put status result for r_token=" << r_token <<
			", to_surl='" << reqfiles.to_surl <<
			"' => status=" << PrintStatus(ReqStatus));
		if (ReqStatus == DPM_READY)
			parse_turl(filestatuses[0].turl);
	} else {
		ReqStatus = DPM_FAILED | EINVAL;
		TRACE(write,"No file statuses returned");
	}

	if (ReqStatus != DPM_QUEUED && ReqStatus != DPM_READY &&
			ReqStatus != DPM_ACTIVE) {
		if ((ReqStatus&0xf000) == DPM_FAILED &&
				DPMRETRY(ReqStatus&0xfff)) {
			// perhaps next time the request will work
			TRACE(write,"Can be retried");
			return;
		}
		if (dpmerr.GetSerrno()) {
			throw DpmFileRequestException(ReqStatus,
				dpmerr.GenerateBrief(),dpmerr.GenerateDetail());
		} else {
			throw DpmFileRequestException(ReqStatus);
		}
	}
}

void DpmFileRequest::do_first_get() {
	EPNAME("DpmFileRequest::do_first_get");
	struct dpm_getfilestatus *filestatuses;
	struct dpm_getfilereq reqfiles;
	char *pp,protocols[32],u_token[256];
	time_t retrytime = 0;
	int nbreplies;
	DpmError dpmerr(epname);

	get_helper h(&reqfiles, &filestatuses, &nbreplies);
	reqfiles.from_surl = strdup(SafeCStr(spath));
	reqfiles.lifetime = ReqOpts.lifetime;
 	reqfiles.f_type = ReqOpts.ftype;

	strcpy(reqfiles.s_token, ReqOpts.stoken);
	// nb protocols array is fixed size
	strcpy(protocols, XPROTO);
	// nb u_token is fixed size
	strcpy(u_token,"DPM xrootd-Finder get request");

	DEBUG("calling with from_surl='" << reqfiles.from_surl <<
		"', lifetime=" << reqfiles.lifetime << ", f_type='" <<
		((reqfiles.f_type) ? XrdOucString(reqfiles.f_type) :
			XrdOucString("")) <<
		"', s_token='" << reqfiles.s_token << "'");

	ReqStatus = 0;
	pp = protocols;
	if (dpm_get(1, &reqfiles, 1, &pp, u_token, retrytime, r_token,
			&nbreplies, &filestatuses)<0) {
		dpmerr.save("dpm_get");
		ReqStatus = DPM_FAILED | serrno;
		TRACE(read,"dpm_get failed with from_surl='"
			<< reqfiles.from_surl << "' => serrno=" << serrno
			<< " detail=" << dpmerr.GenerateDetail());
		if (filestatuses && filestatuses[0].errstring)
			dpmerr.assign(filestatuses[0].errstring);
	} else if (nbreplies) {
		ReqStatus = filestatuses[0].status;
		TRACE(read,"get result for from_surl='" <<
			reqfiles.from_surl << "' => r_token=" << r_token <<
			", status=" << PrintStatus(ReqStatus));
		if (ReqStatus == DPM_READY)
			parse_turl(filestatuses[0].turl);
	} else {
		ReqStatus = DPM_FAILED | EINVAL;
		TRACE(read,"No file statuses returned");
	}

	if (ReqStatus != DPM_QUEUED && ReqStatus != DPM_READY &&
			ReqStatus != DPM_ACTIVE) {
		if ((ReqStatus&0xf000)==DPM_FAILED &&
				DPMRETRY(ReqStatus&0xfff)) {
			// perhaps next time the request will work
			TRACE(read,"Can be retried");
			return;
		}
		if (dpmerr.GetSerrno()) {
			throw DpmFileRequestException(ReqStatus,
				dpmerr.GenerateBrief(),dpmerr.GenerateDetail());
		} else {
			throw DpmFileRequestException(ReqStatus);
		}
	}
	isFirst = 0;
}

void DpmFileRequest::do_getstatus() {
	EPNAME("DpmFileRequest::do_getstatus");
	struct dpm_getfilestatus *filestatuses;
	struct dpm_getfilereq reqfiles;
	int nbreplies;
	DpmError dpmerr(epname);

	ReqStatus = 0;

	get_helper h(&reqfiles, &filestatuses, &nbreplies);
	reqfiles.from_surl = strdup(SafeCStr(spath));

	DEBUG("calling with r_token=" << r_token << ", from_surl='" <<
		reqfiles.from_surl << "'");

	if (dpm_getstatus_getreq(r_token, 1, &reqfiles.from_surl, &nbreplies,
			&filestatuses)<0) {
		dpmerr.save("dpm_getstatus_getreq");
		ReqStatus = DPM_FAILED | serrno;
		TRACE(read,"dpm_getstatus_getreq failed with r_token="
			<< r_token << ", from_surl='" << reqfiles.from_surl
			<< "' => serrno=" << serrno
			<< " detail=" << dpmerr.GenerateDetail());
		if (filestatuses && filestatuses[0].errstring)
			dpmerr.assign(filestatuses[0].errstring);
	} else if (nbreplies) {
		ReqStatus = filestatuses[0].status;
		TRACE(read,"getstatus result for r_token=" << r_token <<
			", from_surl='" << reqfiles.from_surl <<
			"' => status=" << PrintStatus(ReqStatus));
		if (ReqStatus == DPM_READY)
			parse_turl(filestatuses[0].turl);
	} else {
		ReqStatus = DPM_FAILED | EINVAL;
		TRACE(read,"No file statuses returned");
	}

	if (ReqStatus != DPM_QUEUED && ReqStatus != DPM_READY &&
			ReqStatus != DPM_ACTIVE) {
		if ((ReqStatus&0xf000) == DPM_FAILED &&
				DPMRETRY(ReqStatus&0xfff)) {
			// perhaps next time the request will work
			TRACE(read,"Can be retried");
			return;
		}
		if (dpmerr.GetSerrno()) {
			throw DpmFileRequestException(ReqStatus,
				dpmerr.GenerateBrief(),dpmerr.GenerateDetail());
		} else {
			throw DpmFileRequestException(ReqStatus);
		}
	}
}

bool DpmFileRequest::isAccessiblePutOngoing() {
	EPNAME("DpmFileRequest::isAccessiblePutOngoing");
	DpmError dpmerr(epname);

	struct dpns_filereplicax *rep_entries = 0;
	int nbreplicas = 0;

	if (dpns_access(SafeCStr(spath), R_OK|W_OK)<0 ||
			dpns_getreplicax(SafeCStr(spath), NULL, NULL,
				&nbreplicas, &rep_entries)<0) {
		return 0;
	}

	if (nbreplicas != 1 || rep_entries->status != 'P') {
		free(rep_entries);
		return 0;
	}

	char *turl = (char*)malloc(
		strlen(XPROTO) +
		strlen(rep_entries->host) + 4 );

	strcpy(turl,XPROTO);
	strcat(turl,"://");
	strcat(turl,rep_entries->host);

	free(rep_entries);

	try {
		parse_turl(turl);
	} catch (DpmFileRequestException &e) {
		free(turl);
		return 0;
	}

	free(turl);
	ReqStatus = DPM_READY;
	r_token[0] = '\0';
	return 1;
}

void DpmFileRequest::parse_turl(const char *turl) {
	if (!turl)
		throw DpmFileRequestException((ReqStatus&0xf000)|EINVAL,
			"No turl available");
	r_host.assign(turl,0);
	int isl = r_host.find(":/");
	if (isl == STR_NPOS)
		throw DpmFileRequestException((ReqStatus&0xf000)|EINVAL,
			"Could not find ':/' in turl");
	r_host.erasefromstart(isl+2);
	while(r_host.beginswith('/'))
		r_host.erasefromstart(1);
	r_sfn = r_host;
	isl = r_host.find('/');
	if (isl != STR_NPOS) {
		r_host.erase(isl);
		r_sfn.erasefromstart(isl);
	} else {
		r_sfn.erase();
	}
	if (r_host.length() && r_host.endswith(':'))
		r_host.erasefromend(1);
	else if (r_sfn.beginswith('/'))
		r_sfn.erasefromstart(1);
	if (r_host.length() == 0)
		throw DpmFileRequestException((ReqStatus&0xf000)|EINVAL,
			"Invalid hostname portion of turl");
}

/******************************************************************************/
/*                             D F Q u e u e K e y                            */
/******************************************************************************/

DFQueueKey::DFQueueKey(const XrdOucString &path, const XrdOucString &Dn,
	const XrdOucString &Groups, const DpmFileRequestOptions &ReqOpts)
	: path(path), Dn(Dn), Groups(Groups), ReqOpts(ReqOpts) {}

DFQueueKey::~DFQueueKey() {}

bool DFQueueKey::operator==(const DFQueueKey& rhs) const {
	if (!path.length() || !Dn.length())
		return 0;
	if (!rhs.path.length() || !rhs.Dn.length())
		return 0;
	if (strcmp(SafeCStr(path), SafeCStr(rhs.path)) ||
	    strcmp(SafeCStr(Dn), SafeCStr(rhs.Dn)) || 
	    strcmp(SafeCStr(Groups), SafeCStr(rhs.Groups)))
		return 0;
	if (ReqOpts != rhs.ReqOpts)
		return 0;
	return 1;
}

bool DFQueueKey::operator!=(const DFQueueKey& rhs) const
	{ return !(*this == rhs); }

/******************************************************************************/
/*                           D F Q u e u e I t e m                            */
/******************************************************************************/

DFQueueItem::DFQueueItem(const DFQueueKey &key)
	: next(0), key(new DFQueueKey(key)), create_time(time(0)),
	skip_until(0), req(0), tag(0) {}

DFQueueItem::~DFQueueItem() { if (req) delete req; }

DpmFileRequest *DFQueueItem::GetReq() { return req; }

void DFQueueItem::SetReq(DpmFileRequest *req) { this->req = req; }

time_t DFQueueItem::CreateTime() const { return create_time;}

void DFQueueItem::SetSkipDelay(time_t delay) {skip_until = time(0) + delay;}

/******************************************************************************/
/*                        D F Q u e u e A n c h o r                           */
/******************************************************************************/

DFQueueAnchor::DFQueueAnchor() : seq(0) {}

DFQueueAnchor::~DFQueueAnchor() {
	DFQueueItem *p, *q = first;
	while( (p=q) != 0 ) {
		q = q->next;
		p->next = 0;
		delete p;
	}
}

DFQueueItem *DFQueueAnchor::Remove(const DFQueueKey &key) {
	EPNAME("DFQueueAnchor::Remove");
	time_t t0;
	int count=0,del=0;
	DFQueueItem *found=0;
	t0 = time(0);
	XrdOucString removed_tags;

	do {
		XrdSysMutexHelper  hlpMtx(&myMutex);
		DFQueueItem *q = first, *prev=0, *found_prev=0;

		while(q) {
			if ((q->skip_until == 0 || t0 >= q->skip_until) &&
					*q->key == key) {
				found_prev=prev;
				found=q;
			}
			prev = q;
			q = q->next;
			count++;
		}
		if (found) {
			if (found_prev)
				found_prev->next = found->next;
			else
				first = found->next;
			found->next = 0;
		}
		q = first; prev = 0;
		while(q) {
			if (abs(t0 - q->create_time) > MAX_QLIFE) {
				DFQueueItem *tmp=q;
				q = q->next;
				if (prev)
					prev->next = q;
				else
					first = q;
				if (removed_tags.length())
					removed_tags += ",";
				removed_tags += tmp->tag;
				delete tmp;
				del++;
			} else {
				prev = q;
				q = q->next;
			}
		}
	} while(0);

	DEBUG("Queue length " << count << ". " <<
		(del ? ("Deleted old entries [" + removed_tags + "]. ") :
			XrdOucString("")) <<
		"Target entry " <<
		(found ? (XrdOucString("[") + found->tag + "] ") :
			XrdOucString("")) <<
		(found ? "found and removed." : "not found in the queue."));
	return found;
}

void DFQueueAnchor::Insert(DFQueueItem *item) {
	EPNAME("DFQueueAnchor::Insert");
	XrdOucString id;
	do {
		XrdSysMutexHelper  hlpMtx(&myMutex);
		id = (item->tag = ++seq);
		item->next = first;
		first = item;
	} while(0);
	DEBUG("Inserted item [" << id << "]");
}

DFQueueAnchor TheQ;

/******************************************************************************/
/*              D p m F i l e R e q u e s t O p t i o n s                     */
/******************************************************************************/

DpmFileRequestOptions::DpmFileRequestOptions(bool isput, XrdOucEnv *env,
	const DpmFinderConfigOptions &defaults) : isPut(isput)
{
	EPNAME("DpmFileRequestOptions::DpmFileRequestOptions");
	const char *p;

	if (isPut) {
		lifetime = defaults.reqput_lifetime;
		ftype = defaults.reqput_ftype;
		strcpy(stoken, defaults.reqput_stoken);
		reqsize = defaults.reqput_reqsize;
		if (env && (p = env->Get("dpm.reqsize")) != 0) {
			long long Size;
			if (!XrdOuca2x::a2sz(Say,"invalid reqsize",p,&Size,0)) {
				reqsize = Size;
			} else {
				throw DpmFileRequestException(DPM_FAILED|EINVAL,
					"Could not read reqsize in request");
			}
		} else if (env && (p = env->Get("oss.asize")) != 0) {
			long long cgSize;
			if (!XrdOuca2x::a2sz(Say,"invalid asize",p,&cgSize,0)) {
				reqsize = cgSize;
			}
		}
	} else {
		lifetime = defaults.reqget_lifetime;
		ftype = defaults.reqget_ftype;
		strcpy(stoken, defaults.reqget_stoken);
		reqsize = 0;
	}

	if (env && (p = env->Get("dpm.ftype")) != 0)
		ftype = (*p == '-') ? '\0' : *p;

	if (env && (p = env->Get("dpm.stoken")) != 0) {
		if (strlen(p) <= CA_MAXDPMTOKENLEN) {
			strcpy(stoken, p);
		} else {
			throw DpmFileRequestException(DPM_FAILED|ENAMETOOLONG,
				"Requested space token too long");
		}
	}

	if (env && (p = env->Get("dpm.lifetime")) != 0) {
		char *eP;
		lifetime = strtol(p, &eP, 10);
		if (!*p || *eP) {
			throw DpmFileRequestException(DPM_FAILED|EINVAL,
				"Could not read lifetime in request");
		}
	}

	isMkpath = defaults.mkpath_bool;
}

DpmFileRequestOptions::~DpmFileRequestOptions() {}

bool DpmFileRequestOptions::operator==(const DpmFileRequestOptions &rhs) const {
	if (isPut != rhs.isPut)
		return 0;
	if (lifetime != rhs.lifetime || ftype != rhs.ftype ||
			strcmp(stoken, rhs.stoken))
		return 0;
	if (isPut) {
		if (reqsize != rhs.reqsize)
			return 0;
	}
	return 1;
}

bool DpmFileRequestOptions::operator!=(const DpmFileRequestOptions &rhs) const
	{ return !(*this == rhs); }

/******************************************************************************/
/*                 D p m F i n d e r C o n f i g O p t i o n s                */
/******************************************************************************/

DpmFinderConfigOptions::DpmFinderConfigOptions() {
	xrd_server_port  =    0;
	mkpath_bool	 =    1;

	reqput_lifetime  =    0;
	reqput_ftype     =  'P';
	reqput_stoken[0] = '\0';
	reqput_reqsize   =    0;
	reqget_lifetime  =    0;
	reqget_ftype     = '\0';
	reqget_stoken[0] = '\0';

	authorize_bool   =    0;
	gracetime        =    0;

	RedirConfig.theN2N = 0;
}

/******************************************************************************/
/*                          X r d D P M F i n d e r                           */
/******************************************************************************/

XrdDPMFinder::XrdDPMFinder(XrdCmsClient *cmsClient, XrdSysLogger *lp,
	int whoami, int Port) : XrdCmsClient(XrdCmsClient::amRemote),
	Authorization(0), AuthSecondary(0), defaultCmsClient(cmsClient)
{
	Say.logger(lp);
}

XrdDPMFinder::~XrdDPMFinder() {}

int XrdDPMFinder::Configure(const char *cfn, char *Args, XrdOucEnv *EnvInfo) {
	EPNAME("XrdDPMFinder::Configure");
	int cfgFD;
	int NoGo = 0;
	int retc,v;
	const char *var, *val;
	XrdOucEnv myEnv;
	XrdOucStream CFile(&Say, getenv("XRDINSTANCE"), &myEnv, "=====> ");
	DpmFinderConfigOptions *Opts = &DefaultConfig;
	bool enableCms = 0;
	std::vector<XrdOucString> reqhost;

	// copy the ofs trace level
	//
	Trace.What = OfsTrace.What;

	if (!cfn || !*cfn) {
		Say.Emsg("Configure","Configuration file not specified.");
		return 0;
	}

	// Try to open the configuration file.
	//
	if ((cfgFD = open(cfn, O_RDONLY, 0))<0) {
		Say.Emsg("Config", errno, "open config file", cfn);
		return 0;
	}
	CFile.Attach(cfgFD);
	// Start reading records until eof.
	//
	while((var = CFile.GetMyFirstWord())) {
		if (!strncmp(var, "dpm.", 4)) {
			var += 4;

			if (!strcmp("putlifetime",var)) {
				if (!(val = CFile.GetWord())) {
					Say.Emsg("Config",
					"'put' lifetime value not specified.");
					NoGo |= 1;
				} else {
					if(XrdOuca2x::a2tm(Say,
					"putlifetime value", val, &v))
						NoGo |= 1;
					else
						Opts->reqput_lifetime = v;
				}
			} else
			if (!strcmp("putfiletype",var)) {
				if (!(val = CFile.GetWord()) ||
						strlen(val)!=1) {
					Say.Emsg("Config",
			"'put' file type value not specified or invalid.");
					NoGo |= 1;
				} else {
					Opts->reqput_ftype = (*val == '-')
						? '\0' : *val;
				}
			} else
			if (!strcmp("putstoken",var)) {
				if (!(val = CFile.GetWord()) ||
						strlen(val)>CA_MAXDPMTOKENLEN) {
					Say.Emsg("Config",
			"'put' space token not specified or invalid.");
					NoGo |= 1;
				} else { strcpy(Opts->reqput_stoken,val); }
			} else
			if (!strcmp("fixedidrestrict",var)) {
				while((val = CFile.GetWord())) {
					if (!checkDirExists(val)) {
						Say.Emsg("Config",
							"fixedidrestrict path not found");
						NoGo |= 1;
					} else {
						Opts->RedirConfig.AuthLibRestrict.push_back(
							CanonicalisePath(val,1));
					}
				}
			} else
			if (!strcmp("namelib",var)) {
				if (!(val = CFile.GetWord()) || !val[0]) {
					Say.Emsg("Config",
					"namelib not specified.");
					NoGo |= 1;
				} else {
					char parms[1040];
					Opts->N2N_Lib.assign(val,0);
					if (!CFile.GetRest(parms,sizeof(parms))) {
						Say.Emsg("Config","namelib "
						"parameters too long.");
						NoGo |= 1;
					}
					Opts->N2N_Parms.assign(parms,0);
				}
			} else
			if (!strcmp("enablecmsclient",var)) {
				enableCms = 1;
			} else
			if (!strcmp("defaultprefix",var)) {
				if (!(val = CFile.GetWord())) {
					Say.Emsg("Config",
					"'defaultprefix' not specified.");
					NoGo |= 1;
				} else {
					Opts->RedirConfig.defaultPrefix =
						CanonicalisePath(val, 1);
					if (!checkDirExists(Opts->RedirConfig.defaultPrefix)) {
						Say.Emsg("Config",
							"default path prefix not found");
						NoGo |= 1;
					}
				}
			} else
			if (!strcmp("namecheck",var)) {
				while((val = CFile.GetWord())) {
					XrdOucString cs;
					cs = CanonicalisePath(val, 1);
					Opts->RedirConfig.
						N2NCheckPrefixes.push_back(cs);
					if (!checkDirExists(cs)) {
						Say.Emsg("Config",
							"namecheck prefix not found");
						NoGo |= 1;
					}
				}
			} else
			if (!strcmp("allowvo",var)) {
				while((val = CFile.GetWord())) {
					Opts->RedirConfig.IdentConfig.
						validvo.push_back(val);
				}
			} else
			if (!strcmp("localroot",var)) {
				if (!(val = CFile.GetWord())) {
					Say.Emsg("Config",
					"'localroot' not specified.");
					NoGo |= 1;
				} else {
					Opts->RedirConfig.
						lroot_param.assign(val,0);
				}
			} else
			if (!strcmp("replacementprefix",var)) {
				if (!(val = CFile.GetWord())) {
					Say.Emsg("Config",
					"'replacementprefix' initial path "
					"not specified.");
					NoGo |= 1;
				} else {
					XrdOucString pfx1 =
						CanonicalisePath(val, 1);
					if (!(val = CFile.GetWord())) {
						Say.Emsg("Config","'pathprefix'"
						" replacement path not specified.");
						NoGo |= 1;
					} else {
						XrdOucString pfx2 = CanonicalisePath(val, 1);
						Opts->RedirConfig.
							pathPrefixes.push_back(make_pair(pfx1, pfx2));
						if (!checkDirExists(pfx2)) {
							Say.Emsg("Config",
								"replacement path prefix not found");
							NoGo |= 1;
						}
					}
				}
			} else
			if (!strcmp("putsize",var)) {
				if (!(val = CFile.GetWord())) {
					Say.Emsg("Config",
					"defult 'put' size not specified.");
					NoGo |= 1;
				} else {
					if (XrdOuca2x::a2i(Say,
							"putsize value", val, &v))
						NoGo |= 1;
					else
						Opts->reqput_reqsize = v;
				}
			} else
			if (!strcmp("getlifetime",var)) {
				if (!(val = CFile.GetWord())) {
					Say.Emsg("Config", 
					"'get' lifetime value not specified.");
					NoGo |= 1;
				} else {
					if (XrdOuca2x::a2tm(Say,
							"getlifetime value", val, &v))
						NoGo |= 1;
					else
						Opts->reqget_lifetime = v;
				}
			} else
			if (!strcmp("getfiletype",var)) {
				if (!(val = CFile.GetWord()) ||
						strlen(val)!=1) {
					Say.Emsg("Config", 
			"'get' file type value not specified or invalid.");
					NoGo |= 1;
				} else {
					Opts->reqget_ftype = (*val == '-')
						? '\0' : *val;
				}
			} else
			if (!strcmp("getstoken",var)) {
				if (!(val = CFile.GetWord()) ||
						strlen(val)>CA_MAXDPMTOKENLEN) {
					Say.Emsg("Config", 
			"'get' space token not specified or invalid.");
					NoGo |= 1;
				} else { strcpy(Opts->reqget_stoken,val); }
			} else
			if (!strcmp("xrdserverport",var)) {
				if (!(val = CFile.GetWord())) {
					Say.Emsg("Config", 
					"'xrdserverport' is not specified.");
					NoGo |= 1;
				} else {
					if (XrdOuca2x::a2i(Say,
							"xrdserverport value", val, &v))
						NoGo |= 1;
					else
						Opts->xrd_server_port = v;
				}
			} else
			if (!strcmp("mkpath",var)) {
				if (!(val = CFile.GetWord())) {
					Say.Emsg("Config",
					"'mkpath' is not specified.");
					NoGo |= 1;
				} else {
					if (XrdOuca2x::a2i(Say,
							"mkpath value", val, &v))
						NoGo |= 1;
					else
						Opts->mkpath_bool = v;
				}
			} else
			if (!strcmp("mmreqhost",var)) {
				while((val = CFile.GetWord())) {
					reqhost.push_back(val);
				}
			} else
			if (!strcmp("gracetime", var)) {
				if (!(val = CFile.GetWord())) {
					// ok leave the default
					//
				} else {
					if (XrdOuca2x::a2i(Say,
							"gracetime value", val, &v))
						NoGo |= 1;
					else
						Opts->gracetime = v;
				}
			} else
			if (!strcmp("principal",var)) {
				// handled below; need to preserve spaces
			} else
			if (!strcmp("fqan",var)) {
				while((val = CFile.GetWord())) {
					XrdOucString vat, vals(val);
					int from = 0;
					while ((from = vals.tokenize(vat, from, ',')) != -1) {
						if (!vat.length()) continue;
						Opts->RedirConfig.IdentConfig.fqans.push_back(vat);
					}
				}
			} else {
				Say.Emsg("Config", "unknown directive '",var,"'.");
				CFile.Echo();
				NoGo |= 1;
			}

		} else if (!strncmp(var, "ofs.", 4)) {
			var += 4;
			if (!strcmp("authlib",var)) {
				if (!(val = CFile.GetWord())) {
					Say.Emsg("Config",
					"authlib not specified.");
					NoGo |= 1;
				} else {
					Opts->authlib = val;
					char parms[1024];
					if (!CFile.GetRest(parms,sizeof(parms))) {
						Say.Emsg("Config",
						"authlib parameters too long");
						NoGo |= 1;
					} else {Opts->authparm = parms;}
				}
			} else
			if (!strcmp("authorize",var)) {
				Opts->authorize_bool = 1;
			}
		}
	}
	// Check if any errors occured during file i/o
	//
	if ((retc = CFile.LastError()))
		NoGo = Say.Emsg("Config", retc, "read config file", cfn);
	CFile.Close();

	// Try to open the configuration file again for principal
	//
	if ( (cfgFD = open(cfn, O_RDONLY, 0)) < 0) {
		Say.Emsg("Config", errno, "open config file", cfn);
		return 0;
	}
	CFile.Attach(cfgFD);
	while((var = CFile.GetLine())) {
		while(*var==' ' || *var=='\t') var++;
		if (*var=='#') continue;
		if (!strncmp("dpm.principal",var,13)) {
			var += 13;
			if (!(*var==' ' || *var=='\t')) continue;
			while(*var==' ' || *var=='\t') var++;
			const char *tmp = var;
			while(*tmp) tmp++;
			while(tmp!=var && (*(tmp-1)==' '||*(tmp-1)=='\t')) tmp--;
			size_t s = tmp-var;
			if (s>0) {
				Opts->RedirConfig.IdentConfig.principal = XrdOucString(var, s);
			}
		}
	}
	// Now check if any errors occured during file i/o
	//
	if ((retc = CFile.LastError()))
		NoGo = Say.Emsg("Config", retc, "read config file", cfn);
	CFile.Close();

	if (!NoGo && !getenv("XRDDPM_REDIRCONFIG")) {
		char pval[32];
		sprintf(pval,"%p",(void*)&Opts->RedirConfig);
		XrdOucEnv::Export("XRDDPM_REDIRCONFIG",pval);
	}

	// if setting up a cmsclient we expect that the cms will
	// send us stat requests, so prepopulate the allowed host
	// list with this host.
	if (enableCms) {
		std::vector<XrdOucString> lh;
		InitLocalHostNameList(lh);
		reqhost.insert(reqhost.end(),lh.begin(),lh.end());
		reqhost.push_back("localhost");
	}

	// host names to IP addresses
	std::vector<XrdOucString>::const_iterator itr;
	for(itr=reqhost.begin();itr!=reqhost.end();++itr) {
		const char *host = SafeCStr(*itr);
		addrinfo hints;
		addrinfo *ai, *aitop;
		int gaierrno;
		memset (&hints, 0, sizeof(hints));
		hints.ai_family = AF_UNSPEC;
		hints.ai_socktype = SOCK_STREAM;
		gaierrno = getaddrinfo(host, 0, &hints, &aitop);
		if (gaierrno != 0) {
			Say.Emsg("Config","Could not resolve "
				"mmreqhost hostname",host);
			NoGo |= 1;
		} else {
			sockaddr_storage ss;
			for(ai=aitop;ai;ai=ai->ai_next) {
				if (ai->ai_family != AF_INET &&
						ai->ai_family != AF_INET6)
					continue;
				memset(&ss,0,sizeof(ss));
				memcpy(&ss,ai->ai_addr,ai->ai_addrlen);
				Opts->mmReqHosts.push_back(ss);
			}
			freeaddrinfo(aitop);
		}
	}

	if (Opts->key.size() == 0) {
		unsigned char *dat;
		size_t dlen;
		const char *msg;
		if ((msg = LoadKeyFromFile(&dat, &dlen))) {
			Say.Emsg("Config", "Error while reading key from file:",
				msg);
		} else {
			for(size_t idx=0;idx<dlen;++idx)
				Opts->key.push_back(dat[idx]);
			free(dat);
		}
	}
	if (Opts->key.size() < MIN_KEYLEN) {
		char buf[32];
		sprintf(buf, "%d", MIN_KEYLEN);
		Say.Emsg("Config", "The shared key is undefined or too short, "
			"the minimum length required is", buf);
		NoGo |= 1;
	}
	XrdOucString err;
	if (!NoGo && Opts->RedirConfig.IdentConfig.principal.length() &&
		DpmIdentity::badPresetID(Opts->RedirConfig.IdentConfig, err)) {
		Say.Emsg("Config","The preset identity has a problem:",
			SafeCStr(err));
		NoGo |= 1;
	}
	if (!NoGo && Opts->RedirConfig.IdentConfig.principal == "root" &&
			Opts->RedirConfig.IdentConfig.fqans.size() != 0) {
		Say.Emsg("Config","dpm.fqan must not be set if dpm.principal "
			"is 'root'");
		NoGo |= 1;
	}
	if (!NoGo && !Opts->RedirConfig.IdentConfig.principal.length() &&
			Opts->RedirConfig.IdentConfig.fqans.size() != 0) {
		Say.Emsg("Config","Setting dpm.fqan without dpm.principal "
			"has no effect");
		NoGo |= 1;
	}
	// this will set AuthSecondary
	if (!NoGo && Opts->authorize_bool) {
		if (setupAuth(Say.logger(), cfn)) {
			Say.Emsg("Config",
				"Problem configuring authorization library");
			NoGo |= 1;
		}
	}
	if (!NoGo &&
		!(Opts->RedirConfig.IdentConfig.principal.length() == 0 &&
			!AuthSecondary &&
			Opts->RedirConfig.AuthLibRestrict.size() == 0) &&
		!(Opts->RedirConfig.IdentConfig.principal.length() != 0 &&
			AuthSecondary &&
			Opts->RedirConfig.AuthLibRestrict.size() != 0)) {
		Say.Emsg("Config", "dpm.principal, dpm.fixedidrestrict and a "
			"secondary authlib are either all required or should "
			"all be unset.");
		NoGo |= 1;
	}
	if (!NoGo && !Authorization) {
		NoGo |= 1;
		Say.Emsg("Config",
			"DPMFinder requires authorization to be enabled");
	}
	if (!NoGo && Opts->N2N_Lib.length() &&
			(Opts->RedirConfig.pathPrefixes.size()>0 ||
			Opts->RedirConfig.defaultPrefix.length())) {
		Say.Emsg("Config",
			"Can not configure defaultprefix or replacementprefix "
			"with a namelib");
		NoGo |= 1;
	}
	if (!NoGo && !Opts->N2N_Lib.length() &&
			Opts->RedirConfig.lroot_param.length()) {
		Say.Emsg("Config",
			"The localroot parameter has not effect "
			"without a namelib");
		NoGo |= 1;
	}
	if (!NoGo && !Opts->N2N_Lib.length() &&
			Opts->RedirConfig.N2NCheckPrefixes.size()) {
		Say.Emsg("Config",
			"Can not set namecheck without a namelib");
		NoGo |= 1;
	}
	if (!NoGo && Opts->N2N_Lib.length() &&
			!Opts->RedirConfig.N2NCheckPrefixes.size()) {
		Say.Emsg("Config",
			"Must set namecheck when using a namelib");
		NoGo |= 1;
	}
	if (!NoGo && Opts->N2N_Lib.length()) {
		if (setupN2N(Say, cfn)) {
			Say.Emsg("Config",
				"Problem configuring namelib");
			NoGo |= 1;
		}
	}
	if (enableCms) {
		// start standard cms client
		//
		if (!defaultCmsClient ||
		!defaultCmsClient->Configure(cfn, Args, EnvInfo)) {
			Say.Emsg("Configure","Could not configure cms client");
			NoGo |= 1;
		}
	}

	if (NoGo) {
		Say.Say("Error while processing configuration for "
			"DPM Finder plugin");
	} else {
		for(std::vector<XrdOucString>::const_iterator itr=
				Opts->RedirConfig.IdentConfig.validvo.begin();
				itr!=Opts->RedirConfig.IdentConfig.validvo.end();
				++itr) {
			TRACEX("Allowed vo       = " << *itr);
		}
		if (!Opts->RedirConfig.IdentConfig.validvo.size()) {
			TRACEX("Allowed vo       = <any>");
		}
		if (Opts->xrd_server_port) {
			TRACEX("xrd_server_port  = " << Opts->xrd_server_port);
		} else {
			TRACEX("xrd_server_port  = <xroot default>");
		}
		TRACEX("mkpath_bool      = " << (Opts->mkpath_bool
			? "on" : "off"));
		if (Opts->reqput_lifetime) {
			TRACEX("reqput_lifetime  = " << Opts->reqput_lifetime);
		} else {
			TRACEX("reqput_lifetime  = <dpm default>");
		}
		if (Opts->reqput_ftype) {
			TRACEX("reqput_ftype     = '" <<
				Opts->reqput_ftype << "'");
		} else {
			TRACEX("reqput_ftype     = <dpm default>");
		}
		if (*Opts->reqput_stoken) {
			TRACEX("reqput_stoken    = '" <<
				Opts->reqput_stoken << "'");
		} else {
			TRACEX("reqput_stoken    = <none>");
		}
		if (Opts->reqput_reqsize) {
			TRACEX("reqput_reqsize   = " << Opts->reqput_reqsize);
		} else {
			TRACEX("reqput_reqsize   = <dpm default>");
		}
		if (Opts->reqget_lifetime) {
			TRACEX("reqget_lifetime  = " << Opts->reqget_lifetime);
		} else {
			TRACEX("reqget_lifetime  = <dpm default>");
		}
		if (Opts->reqget_ftype) {
			TRACEX("reqget_ftype     = '" <<
				Opts->reqget_ftype << "'");
		} else {
			TRACEX("reqget_ftype     = <dpm default>");
		}
		if (*Opts->reqget_stoken) {
			TRACEX("reqget_stoken    = '" <<
				Opts->reqget_stoken << "'");
		} else {
			TRACEX("reqget_stoken    = <none>");
		}
		if (Opts->RedirConfig.IdentConfig.principal.length()) {
			TRACEX("principal        = '" <<
				Opts->RedirConfig.IdentConfig.principal << "'");
		} else {
			TRACEX("principal        = <none>");
		}
		for(std::vector<XrdOucString>::const_iterator itr=
				Opts->RedirConfig.IdentConfig.fqans.begin();
				itr != Opts->RedirConfig.IdentConfig.fqans.end();
				++itr) {
			TRACEX("fqan             = " << *itr);
		}
		if (Opts->gracetime) {
			TRACEX("max token life   = " << Opts->gracetime
				<< " seconds");
		} else {
			TRACEX("max token life   = <none specified, disk "
				"server discretion>");
		}
		if (AuthSecondary) {
			TRACEX("fixed ID authz   = Yes");
		} else {
			TRACEX("fixed ID authz   = No");
		}
		for(std::vector<sockaddr_storage>::const_iterator itr=
				Opts->mmReqHosts.begin();
				itr!=Opts->mmReqHosts.end();++itr) {
			union {
				sockaddr_storage ss;
				sockaddr sa;
			} u;
			u.ss = *itr;
			TRACEX("mmReqHost        = "
				<< SockaddrToIP(&u.sa));
		}
		if (enableCms) {
			TRACEX("CmsClient (TRG)  = Yes");
		} else {
			TRACEX("CmsClient (TRG)  = No");
		}
		if (Opts->RedirConfig.theN2N) {
			TRACEX("N2N              = " << Opts->N2N_Lib);
		} else {
			TRACEX("N2N              = <none>");
		}
		for(std::vector<XrdOucString>::const_iterator itr=
				Opts->RedirConfig.N2NCheckPrefixes.begin();
				itr!=Opts->RedirConfig.N2NCheckPrefixes.end();
				++itr) {
			TRACEX("Allowed name pfx = " << *itr);
		}
		if (Opts->RedirConfig.lroot_param.length()) {
			TRACEX("N2N LocalRoot    = "
				<< Opts->RedirConfig.lroot_param);
		} else if (Opts->RedirConfig.theN2N) {
			TRACEX("N2N LocalRoot    = <none>");
		}
		for(std::vector<std::pair<XrdOucString, XrdOucString > >::const_iterator itr=
				Opts->RedirConfig.pathPrefixes.begin();
				itr!=Opts->RedirConfig.pathPrefixes.end();
				++itr) {
			TRACEX("Replace prefix   = "
				<< itr->first << " by " << itr->second);
		}
		if (!Opts->RedirConfig.theN2N &&
				!Opts->RedirConfig.pathPrefixes.size()) {
			TRACEX("Replace prefix   = <none>");
		}
		if (Opts->RedirConfig.defaultPrefix.length()) {
			TRACEX("Default prefix   = "
				<< Opts->RedirConfig.defaultPrefix);
		} else if (!Opts->RedirConfig.theN2N) {
			TRACEX("Default prefix   = <none>");
		}
		Say.Say("DPM Finder plugin initialisation completed.");
	}

	// Return final return code
	//
	return NoGo ? 0 : 1;
}

int XrdDPMFinder::Forward(XrdOucErrInfo &Resp, const char *cmd,
                       const char *arg1,  const char *arg2,
                       XrdOucEnv  *Env1,  XrdOucEnv  *Env2)
{
	EPNAME("XrdDPMFinder::Forward");
	std::auto_ptr<DpmIdentity> identP;
	try {
		identP.reset(new DpmIdentity(Env1,
				DefaultConfig.RedirConfig.IdentConfig));
		identP->SetAuth();
	} catch(DpmIdentityException &e) {
		Resp.setErrInfo(e.ErrNum(), e.ErrStr());
		XrdOucString err =
			XrdOucString("Authentication error: ")
			+ e.DetailStr();
		Say.Emsg("Forward", SafeCStr(err));
		return SFS_ERROR;
	}

	DpmError dpmerr(epname);
	bool isTwoWay;
	if ((isTwoWay = (*cmd == '+'))) cmd++;

	XrdOucString FullPath1;
	try {
		FullPath1 = TranslatePath(DefaultConfig.RedirConfig,arg1);
	} catch(int e) {
		XrdOucString err = "Name error: ";
		err += XrdOucString(arg1) + " (" + sstrerror(e) + ")";
		Resp.setErrInfo(e, SafeCStr(err));
		DEBUG(err);
		return SFS_ERROR;
	}

	DEBUG("Processing forward command=" << cmd <<
		" for " << FullPath1 << " isTwoWay=" << isTwoWay);

	if (!strcmp(cmd, "chmod")) {
		mode_t mode;
		TRACE(chmod,"chmod " << arg1 << " to mode " << arg2);
		if (!getMode(arg2, mode)) {
			Resp.setErrInfo(EINVAL, "invalid mode");
			return SFS_ERROR;
		}
		if (dpns_chmod(SafeCStr(FullPath1), mode)<0) {
			dpmerr.save("dpns_chmod");
			DEBUG(dpmerr.GenerateDetail());
			Resp.setErrInfo(dpmerr.GetSerrno(),dpmerr.GenerateBrief());
			if (isTwoWay && DPMRETRY(dpmerr.GetSerrno()))
				return META_RETRYTIME;
			return SFS_ERROR;
		}
		return 0;
	}
	if (!strcmp(cmd, "mkpath") || !strcmp(cmd,"mkdir")) {
		mode_t mode;
		TRACE(mkdir,cmd << " " << arg1 << " mode " << arg2);
		int retc;
		if (!getMode(arg2, mode)) {
			Resp.setErrInfo(EINVAL, "invalid mode");
			return SFS_ERROR;
		}
		if (!strcmp(cmd,"mkpath")) {
			if ((retc=mkp(SafeCStr(FullPath1), mode))>0) {
				Resp.setErrInfo(retc, "error during make path");
				if (isTwoWay && DPMRETRY(retc))
					return META_RETRYTIME;
				return SFS_ERROR;
			}
		}
		if (dpns_mkdir(SafeCStr(FullPath1),mode)<0) {
			dpmerr.save("dpns_mkdir");
			DEBUG(dpmerr.GenerateDetail());
			Resp.setErrInfo(dpmerr.GetSerrno(),dpmerr.GenerateBrief());
			if (isTwoWay && DPMRETRY(dpmerr.GetSerrno()))
				return META_RETRYTIME;
			return SFS_ERROR;
		}
		return 0;
	}
	if (!strcmp(cmd, "rm")) {
		TRACE(remove,"rm " << arg1);
		char *pp = strdup(SafeCStr(FullPath1));
		int nbreplies = 0;
		struct dpm_filestatus *filestatuses=0;
		int retc;
		if ((retc = dpm_rm(1, &pp, &nbreplies, &filestatuses))<0) {
			dpmerr.save("dpm_rm");
			DEBUG(dpmerr.GenerateDetail());
			Resp.setErrInfo(dpmerr.GetSerrno(),dpmerr.GenerateBrief());
		}
		free(pp);
		dpm_free_filest(nbreplies, filestatuses);
		if (retc<0) {
			if (isTwoWay && DPMRETRY(dpmerr.GetSerrno()))
				return META_RETRYTIME;
			return SFS_ERROR;
		}
		return 0;
	}
	if (!strcmp(cmd, "rmdir")) {
		TRACE(remove,"rmdir " << arg1);
		if (dpns_rmdir(SafeCStr(FullPath1))<0) {
			dpmerr.save("dpns_rmdir");
			DEBUG(dpmerr.GenerateDetail());
			Resp.setErrInfo(dpmerr.GetSerrno(),dpmerr.GenerateBrief());
			if (isTwoWay && DPMRETRY(dpmerr.GetSerrno()))
				return META_RETRYTIME;
			return SFS_ERROR;
		}
		return 0;
	}
	if (!strcmp(cmd, "mv")) {
		TRACE(rename,"rename " << arg1 << " to " << arg2);
		XrdOucString FullPath2;
		try {
			FullPath2 = TranslatePath(
				DefaultConfig.RedirConfig, arg2);
		} catch(int e) {
			XrdOucString err = "Name error: ";
			err += XrdOucString(arg2) + " (" + sstrerror(e) + ")";
			Resp.setErrInfo(e, SafeCStr(err));
			DEBUG(err);
			return SFS_ERROR;
		}
		if (dpns_rename(SafeCStr(FullPath1),SafeCStr(FullPath2))<0) {
			dpmerr.save("dpns_rename");
			DEBUG(dpmerr.GenerateDetail());
			Resp.setErrInfo(dpmerr.GetSerrno(),dpmerr.GenerateBrief());
			if (isTwoWay && DPMRETRY(dpmerr.GetSerrno()))
				return META_RETRYTIME;
			return SFS_ERROR;
		}
		return 0;
	}

	DEBUG("Command not supported");
	Resp.setErrInfo(ENOTSUP, "Command not supported");
	return SFS_ERROR;
}

int XrdDPMFinder::Locate(XrdOucErrInfo &Resp, const char *path, int flags,
                      XrdOucEnv *Info)
{
	EPNAME("XrdDPMFinder::Locate");

	DEBUG("Entering Locate, this=" << this << " name="
		<< Resp.getErrUser() << " flags=" << std::hex
		<< flags << std::dec);

	if (flags & SFS_O_META) {
		// This should have been sent to Forward()
		//
		XrdOucString err =
			"Metadata request sent to the "
			"wrong function. Check ofs.forward directive "
			"enables forwarding for all requests.";
		Say.Emsg("Locate", SafeCStr(err));
		Resp.setErrInfo(EINVAL, "Unexpected request");
		return SFS_ERROR;
	}

	bool isCreate = flags & (SFS_O_CREAT | SFS_O_TRUNC);
	bool isPut = isCreate || (flags & (SFS_O_WRONLY | SFS_O_RDWR));
	bool isForOSS = flags & (SFS_O_STAT | SFS_O_LOCATE);

	// is it a stat to discover a file,
	// triggered by a proxy xrootd redirector-cmsd at our site?
	// in either case get and set relevant identity
	//
	bool isXPList = 0;
	const char *apath = path;
	std::auto_ptr<DpmIdentity> identP;

	if ((flags & SFS_O_LOCATE) && *path == '*') {
		apath++;
		if (!*apath) { isXPList=1; }
	}

	if (isForOSS && 
		IsMetaManagerDiscover(Info, flags, path, Resp.getErrUser()))
		{
		try {
			identP.reset(new DpmIdentity());
			identP->SetAuth();
		} catch(DpmIdentityException &e) {
			Resp.setErrInfo(e.ErrNum(), e.ErrStr());
			XrdOucString err =
				XrdOucString(
				"Problem while setting up root identity:  "
				) + e.DetailStr();
			Say.Emsg("Locate", SafeCStr(err));
			return SFS_ERROR;
		}
		TRACE(exists,"Discovery request for " << path);
	} else {
		try {
			identP.reset(new DpmIdentity(Info,
				DefaultConfig.RedirConfig.IdentConfig));
			identP->SetAuth();
		} catch(DpmIdentityException &e) {
			Resp.setErrInfo(e.ErrNum(), e.ErrStr());
			XrdOucString err =
				XrdOucString("Authentication error: ")
				+ e.DetailStr();
			Say.Emsg("Locate", SafeCStr(err));
			return SFS_ERROR;
		}

		// do an authorizaiton check directly to the secondary;
		// to cover cases where the Ofs has not yet already
		// called Authorize or did not also provie the opaque
		// information, or for AOP_Stat the secondary was bypassed.
		if (DpmIdentity::usesPresetID(Info)) {
			Access_Operation ao = AOP_Read;
			const char *action = "open";
			if (isForOSS) {
				ao = AOP_Stat;
				action = "locate";
			} else if (isCreate) {
				ao = AOP_Create;
				action = "create";
			} else if (isPut) {
				ao = AOP_Update;
			}
			const XrdSecEntity *secEntity = Info->secEnv();
			if (!AuthSecondary || (!isXPList &&
					!Authorization->Access(secEntity, apath,
					ao, (flags & SFS_O_LOCATE)? 0: Info))) {
				char buffer[MAXPATHLEN+80];
				if (!AuthSecondary) {
					DEBUG("Use of fixed id needs a "
					"secondary authorization library to "
					"be configured. Denying");
				}
				snprintf(buffer, sizeof(buffer),
					"Unable to %s %s; Permission denied",
					action, *apath ? apath : "[none]");
				Say.Emsg(epname, Resp.getErrUser(), buffer);
				Resp.setErrInfo(EACCES, buffer);
				return SFS_ERROR;
			}
		}
		if (flags & SFS_O_STAT) {
			TRACE(exists, "Stat request for " << path);
		} else if (flags & SFS_O_LOCATE) {
			TRACE(fsctl, "Locate request for " << path);
		}
	}

	XrdOucString FullPath;
	try {
		if (!isXPList) {
			FullPath = TranslatePath(
				DefaultConfig.RedirConfig, apath);
		}
	} catch(int e) {
		XrdOucString err = "Name error: ";
		err += XrdOucString(apath) + " (" + sstrerror(e) + ")";
		Resp.setErrInfo(e, SafeCStr(err));
		DEBUG(err);
		return SFS_ERROR;
	}

	if (isForOSS) {
		// Our Oss has to handle this. The authorization check
		// has already been done either by the Ofs or above.
		if (!Info) {
			Resp.setErrInfo(EINVAL, "No environment");
			return SFS_ERROR;
		}
		Info->Put("dpm.dn",SafeCStr(EncodeString(identP->Dn())));
		Info->Put("dpm.voms",SafeCStr(EncodeString(identP->Groups())));
		Info->Put("dpm.surl",SafeCStr(EncodeString(FullPath)));
		
		// Try to skip an unnecessary lookup to the dpns
		// in the Locate case
		if (flags & SFS_O_LOCATE)
			Info->Put("dpm.skiplocate", "1");
		
		DEBUG("Sending to Oss");
		return 0;
	}

	// needs file access, try to get it and redirect as needed
	//
	int ret;
	try {
		ret = DoFileAccessRequest(Resp, path, FullPath, flags,
			isPut, *identP, Info);
	} catch(DpmFileRequestException &e) {
		Resp.setErrInfo(e.ErrCode(), e.ErrStr());
		DEBUG("DPM File Request status "
			<< PrintStatus(e.DpmStatus())
			<< " => " << e.DetailStr());
		return SFS_ERROR;
	}
	return ret;
}

int XrdDPMFinder::DoFileAccessRequest(XrdOucErrInfo &Resp,
	const char *xrd_fn, const XrdOucString &FullPath,
	int flags, bool isPut, const DpmIdentity &ident, XrdOucEnv *Info)
{
	EPNAME("XrdDPMFinder::DoFileAccessRequest");

	DpmFileRequestOptions ReqOpts(isPut, Info, DefaultConfig);
	DFQueueKey qk(FullPath, ident.Dn(), ident.Groups(), ReqOpts);

	std::auto_ptr<DFQueueItem> qitem(TheQ.Remove(qk));
	DpmFileRequest *req;
	bool first_query=0;

	if (! qitem.get()) {
		std::auto_ptr<DpmFileRequest> r(new DpmFileRequest(FullPath,
			flags, ReqOpts));
		qitem.reset(new DFQueueItem(qk));
		qitem->SetReq(r.get());
		req = r.release();
		first_query = 1;
	} else {
		req = qitem->GetReq();
	}

	suseconds_t idelay = 125000;
	struct timeval t_start, t_now, t_elapsed, t_delay;
	gettimeofday(&t_start, 0);
	while(1) {
		req->UpdateStatus();

		if (req->LastStatus() == DPM_READY) {
			XrdOucString targ;
			try {
				targ = req->Host() + "?" +
					mk_opaque(
					xrd_fn,
					*req,
					ident,
					DefaultConfig.key,
					DefaultConfig.gracetime);

			} catch (const char *err) {
				Resp.setErrInfo(EINVAL, "Unable to redirect");
				Say.Emsg("DoFileAccessRequest", err);
				return SFS_ERROR;
			}
			Resp.setErrInfo(DefaultConfig.xrd_server_port,
				SafeCStr(targ));
			int isl = targ.find("&dpm.hv1=");
			if (isl != STR_NPOS) {
				isl += 9;
				int isl2 = targ.find('&',isl);
				if (isl2 == STR_NPOS)
					targ.erase(isl);
				else if (isl2 > isl)
					targ.erase(isl,isl2-isl);
				targ.insert("[removed]",isl);
			}
			TRACE(redirect,"Redirecting to " << targ);
			return SFS_REDIRECT;
		} else if (abs(time(0) - qitem->CreateTime()) > QABORT_LIMIT) {
			try {
				req->Abort();
			} catch(DpmFileRequestException &e) {
				// if the abort fails no further recovery is
				// tried
			}
			Resp.setErrInfo(ETIMEDOUT,
				"Timed out waiting for file access request");
			DEBUG("Timed out, request not making progress");
			return SFS_ERROR;
		}

		if (!first_query || idelay>500000)
			break;

		gettimeofday(&t_now, 0);
		timersub(&t_now, &t_start, &t_elapsed);
		if (t_elapsed.tv_sec != 0)
			break;

		t_delay.tv_sec = 0;
		t_delay.tv_usec = idelay;
		select(0, 0, 0, 0, &t_delay);

		idelay *= 2;
	}

	int waittime = req->BackoffWaitTime();
	qitem->SetSkipDelay(waittime);
	TheQ.Insert(qitem.get());
	qitem.release();

	TRACE(delay,"returning waittime=" << waittime);
	return waittime;
}

int XrdDPMFinder::Space(XrdOucErrInfo &Resp, const char *path, XrdOucEnv *Info)
{
	EPNAME("XrdDPMFinder::Space");
	std::auto_ptr<DpmIdentity> identP;

	if (!Info) {
		Resp.setErrInfo(EINVAL, "No environment");
		return SFS_ERROR;
	}

	// do an authorizaiton check directly to the secondary;
	// to cover cases where the Ofs has not yet already
	// called Authorize or did not also provie the opaque
	// information.
	if (DpmIdentity::usesPresetID(Info)) {
		const XrdSecEntity *secEntity = Info->secEnv();
		if (!AuthSecondary ||
				!Authorization->Access(secEntity, path,
				AOP_Stat, 0)) {
			char buffer[MAXPATHLEN+80];
			if (!AuthSecondary) {
				DEBUG("Use of fixed id needs a secondary "
				"authorization library to be configured. "
				"Denying");
			}
			snprintf(buffer, sizeof(buffer),
				"Unable to statfs %s; Permission denied", path);
			Say.Emsg(epname, Resp.getErrUser(), buffer);
			Resp.setErrInfo(EACCES, buffer);
			return SFS_ERROR;
		}
	}

	try {
		identP.reset(new DpmIdentity(Info,
				DefaultConfig.RedirConfig.IdentConfig));
		identP->SetAuth();
	} catch(DpmIdentityException &e) {
		Resp.setErrInfo(e.ErrNum(), e.ErrStr());
		XrdOucString err =
			XrdOucString("Authentication error: ")
			+ e.DetailStr();
		Say.Emsg("Space", SafeCStr(err));
		return SFS_ERROR;
	}

	XrdOucString FullPath;
	try {
		FullPath = TranslatePath(DefaultConfig.RedirConfig, path);
	} catch(int e) {
		XrdOucString err = "Name error: ";
		err += XrdOucString(path) + " (" + sstrerror(e) + ")";
		Resp.setErrInfo(e, SafeCStr(err));
		DEBUG(err);
		return SFS_ERROR;
	}
	Info->Put("dpm.dn",SafeCStr(EncodeString(identP->Dn())));
	Info->Put("dpm.voms",SafeCStr(EncodeString(identP->Groups())));
	Info->Put("dpm.surl",SafeCStr(EncodeString(FullPath)));
	DEBUG("Sending to Oss, dpm.surl=" << FullPath);
	return 0;
}

// based on XrdOssSys::ConfigN2N
//
int XrdDPMFinder::setupN2N(XrdSysError &Eroute, const char *cfn)
{
	XrdSysPlugin    *myLib;
	typedef XrdOucName2Name *(*ept)(XrdOucgetName2NameArgs);

	if (!(myLib = new XrdSysPlugin(&Eroute,
		SafeCStr(DefaultConfig.N2N_Lib)))) return 1;

	ept ep = reinterpret_cast<ept>(reinterpret_cast<size_t>
			(myLib->getPlugin("XrdOucgetName2Name")));

	if (!ep) return 1;

	DefaultConfig.RedirConfig.theN2N = ep(&Eroute, cfn,
		SafeCStr(DefaultConfig.N2N_Parms),
		DefaultConfig.RedirConfig.lroot_param.length() ? 
			SafeCStr(DefaultConfig.RedirConfig.lroot_param) :
			0,
		0);

	return DefaultConfig.RedirConfig.theN2N == 0;
}

// based on XrdOfs::setupAuth
//
int XrdDPMFinder::setupAuth(XrdSysLogger *lp, const char *cfn)
{
	XrdSysPlugin    *myLib;
	typedef	XrdAccAuthorize *(*ept)(XrdSysLogger *, const char *,
		const char *);

	if (!DefaultConfig.authlib.length())
		return 1;

	if (!(myLib = new XrdSysPlugin(&Say, SafeCStr(DefaultConfig.authlib))))
		return 1;

	ept ep = reinterpret_cast<ept>(reinterpret_cast<size_t>
			(myLib->getPlugin("XrdAccAuthorizeSecondaryObject")));

	if (!ep) return 1;

	if ((Authorization = ep(lp, cfn, SafeCStr(DefaultConfig.authparm)))) {
		AuthSecondary = 1;
		return 0;
	}

	ep = reinterpret_cast<ept>(reinterpret_cast<size_t>
			(myLib->getPlugin("XrdAccAuthorizeObject")));
	
	if (!ep) return 1;

	AuthSecondary = 0;

	return 0 ==
		(Authorization = ep(lp, cfn, SafeCStr(DefaultConfig.authparm)));
}

bool XrdDPMFinder::IsMetaManagerDiscover(XrdOucEnv *Info, int flags,
                        const char *path, const char *user)
{
	EPNAME("XrdDPMFinder::IsMetaManagerDiscover");

	if (!(flags & SFS_O_STAT))
		return 0;
	if (!Info || !Info->Get("oss.lcl"))
		return 0;
	if (!path || !*path)
		return 0;
	if (!user || !*user)
		return 0;

	XrdLinkMatch match;
	{
		char *susr = (char*)malloc(strlen(user)+2);
		strcpy(susr, user);
		strcat(susr, "*");
		match.Set(susr);
		DEBUG("Link match: " << susr);
		free(susr);
	}
        XrdLink *linkp;
	XrdOucString hn,ip;
        int curr = -1, nmatch = 0;
	union {
		sockaddr_storage ss;
		sockaddr_in sin;
		sockaddr_in6 sin6;
		sockaddr sa;
	} u_naddr,u_itr;

	memset(&u_naddr, 0, sizeof(u_naddr));

	// must not exit this loop early, it will leave list locked
	//
        while((linkp = XrdLink::Find(curr, &match))) {
		if (!strcmp(linkp->ID, user)) {
			hn.assign(linkp->Host(&u_naddr.sa), 0);
			++nmatch;
		}
        }

	if (nmatch != 1) {
		DEBUG("Found " << nmatch << " link matches");
		return 0;
	}

	bool hfound=0;
	std::vector<sockaddr_storage>::const_iterator itr;

	in_addr *targ = &u_naddr.sin.sin_addr;
	in6_addr *targ6 = &u_naddr.sin6.sin6_addr;

	ip = SockaddrToIP(&u_naddr.sa);

	for(itr=DefaultConfig.mmReqHosts.begin();
			itr!=DefaultConfig.mmReqHosts.end();++itr) {

		u_itr.ss = *itr;

		in_addr *in = &u_itr.sin.sin_addr;
		in6_addr *in6 = &u_itr.sin6.sin6_addr;

		if (itr->ss_family == AF_INET) {
			if (u_naddr.ss.ss_family == AF_INET6) {
				if (IN6_IS_ADDR_V4MAPPED(targ6) &&
				!memcmp(in, &targ6->s6_addr[12], 4)) {
					hfound=1;
					break;
				}
				if (ntohl(in->s_addr) == INADDR_LOOPBACK &&
						IN6_IS_ADDR_LOOPBACK(targ6)) {
					hfound=1;
					break;
				}
			}
			if (u_naddr.ss.ss_family == AF_INET) {
				if (!memcmp(in, targ, sizeof(in_addr))) {
					hfound=1;
					break;
				}
			}
		}
		if (itr->ss_family == AF_INET6) {
			if (u_naddr.ss.ss_family == AF_INET) {
				if (IN6_IS_ADDR_V4MAPPED(in6) &&
				!memcmp(&in6->s6_addr[12], targ, 4)) {
					hfound=1;
					break;
				}
				if (IN6_IS_ADDR_LOOPBACK(in6) &&
				ntohl(targ->s_addr) == INADDR_LOOPBACK) {
					hfound=1;
					break;
				}
			}
			if (u_naddr.ss.ss_family == AF_INET6) {
				if (!memcmp(in6, targ6,
						sizeof(in6_addr))) {
					hfound=1;
					break;
				}
			}
		}
	}

	if (!hfound) {
		XrdOucString err =
			"Possible meta-manager discovery request from host "
			+ hn + " " + ip
			+ " (not listed in dpm.mmreqhost)";
		Say.Emsg("IsMetaManagerDiscover", SafeCStr(err));
		return 0;
	}

	return 1;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/

extern "C"
{
XrdCmsClient *XrdCmsGetClient(XrdSysLogger *Logger, // Where messages go
                              int           opMode, // Operational mode
                              int           myPort, // Server's port number
                              XrdOss	   *theSS)  // Storage System I/F
{
	XrdCmsClient *cmsFinder = (XrdCmsClient *)new
		XrdCmsFinderTRG(Logger,XrdCms::IsTarget,myPort);

	XrdDPMFinder *inst = new XrdDPMFinder(cmsFinder,Logger,opMode,myPort);

	return (XrdCmsClient *)inst;
}
} // extern "C"
