/* 
 * File:   XrdDPMOss.cc
 * Author: Fabrizio Furano, David Smith (CERN IT/GT)
 * 
 * An Oss plugin that interacts with a DPM
 *
 * Created on December 21, 2011, 10:25 AM
 */

#include "XrdDPM/XrdCompileVersion.hh"
#include "XrdDPM/XrdDPMOss.hh"
#include "XrdOss/XrdOssTrace.hh"
#include "XrdOuc/XrdOucEnv.hh"
#include "XrdSec/XrdSecEntity.hh"

#include <list>

#include <fcntl.h>
#include <sys/types.h>

DFThreadSpecific dpm_sf;

class fListItem {
public:
  	fListItem(XrdDPMOssFile *FileObjectP,
                const char *fn, XrdOucString Sfn) : XrdFileName(fn),
                FileObjectP(FileObjectP), Sfn(Sfn) { };
  	XrdOucString XrdFileName;
  	XrdDPMOssFile *FileObjectP;
        XrdOucString Sfn;
};

static XrdSysMutex fLLock;
static std::list<fListItem> fList;

static void deleteObjFList(const XrdDPMOssFile *obj)
{
	std::list<fListItem>::iterator itr;
	XrdSysMutexHelper hlpMtx(&fLLock);
	itr = fList.begin();
	while(itr != fList.end()) {
		if (itr->FileObjectP == obj) {
			fList.erase(itr++);
		} else {
			++itr;
		}
	}
}

static void getSfnsFromList(
	const XrdDPMOssFile *obj,
	const char *fn,
	std::vector<XrdOucString> &svec)
{
	std::list<fListItem>::iterator itr;
	svec.clear();
	XrdSysMutexHelper hlpMtx(&fLLock);
	for(itr=fList.begin(); itr != fList.end(); ++itr) {
		if ((!obj || itr->FileObjectP == obj) &&
				(!fn || itr->XrdFileName == fn)) {
			svec.push_back(itr->Sfn);
		}
	}
}

/******************************************************************************/
/*                  E r r o r   R o u t i n g   O b j e c t                   */
/******************************************************************************/

XrdOssSys *XrdOssSS = 0;

XrdSysError OssEroute(0, "DPMoss_");

XrdOucTrace OssTrace(&OssEroute);


XrdDPMOssDir::XrdDPMOssDir(const char *tid):
		XrdOssDir(tid), RedirConfig(0), dpns_dirp(0) {
	const char *p = getenv("XRDDPM_REDIRCONFIG");
	if (p) {
		void *p2;
		sscanf(p, "%p", &p2);       
		RedirConfig = (DpmRedirConfigOptions*)p2;    
	}
}

XrdDPMOssDir::~XrdDPMOssDir() {
	long long retsz;

	if (dpns_dirp)
		Close(&retsz);
}

int XrdDPMOssDir::Close(long long *retsz) {
	const char *tident = "tident";
	EPNAME("XrdDPMOssDir::Close");
	int ret;

	if (!dpns_dirp)
		return -8002;

	try {
		identP->SetAuth();
	} catch(DpmIdentityException &e) {
		return -EACCES;
	}

	DpmError dpmerr(epname);
	ret = dpns_closedir(dpns_dirp);
	dpmerr.save("dpns_closedir");

	identP->UnsetAuth();

	if (ret) {
		return -dpmerr.GetSerrno();
	}

	dpns_dirp = 0;

	return 0;
}

int XrdDPMOssDir::Opendir(const char *path, XrdOucEnv &Info) {
	const char *tident = "tident";
	EPNAME("XrdDPMOssDir::Opendir");

	if (dpns_dirp)
		return -8003;

	if (!RedirConfig)
		return -ENOTSUP;

        try {
                identP.reset(new DpmIdentity(&Info, RedirConfig->IdentConfig));
                identP->SetAuth();
        } catch(DpmIdentityException &e) {
		return -EACCES;
        }

        XrdOucString FullPath;
	try {
                FullPath = TranslatePath(*RedirConfig, path);
	} catch(int e) {
		return -e;
        }
	DpmError dpmerr(epname);

	dpns_dirp = dpns_opendir(SafeCStr(FullPath));
	dpmerr.save("dpns_opendir");

	identP->UnsetAuth();

	if (!dpns_dirp) {
		return -dpmerr.GetSerrno();
	}

	return 0;
}

int XrdDPMOssDir::Readdir(char *buff, int blen) {
	const char *tident = "tident";
	EPNAME("XrdDPMOssDir::Readdir");

	if (!dpns_dirp)
		return -8002;

	struct dirent *de;

	try {
		identP->SetAuth();
	} catch(DpmIdentityException &e) {
		return -EACCES;
	}

	DpmError dpmerr(epname);
	de = dpns_readdir(dpns_dirp);
	dpmerr.save("dpns_readdir");

	identP->UnsetAuth();

	if (de) {
		strncpy(buff, de->d_name, blen);
		buff[blen-1] = '\0';
		return 0;
	}

	buff[0] = '\0';
	return -dpmerr.GetSerrno();
}

XrdDPMOssFile::~XrdDPMOssFile() {
	if (mydpmtk) free(mydpmtk);
	mydpmtk = 0;

	if (myfilename) free(myfilename);
	myfilename = 0;

	deleteObjFList(this);
}

int XrdDPMOssFile::Fchmod(mode_t mode) {
	const char *tident = "tident";
	EPNAME("XrdDPMOssFile::Fchmod");

        bool isSuid = mode & S_ISUID;
        int ret = XrdOssFile::Fchmod(0660 | (isSuid ? S_ISUID : 0));

        DEBUG("Issued XrdOssFile::Fchmod to " << 
		(isSuid ? "set" : "clear") << " S_ISUID ret=" << ret);
        
	return ret;
}

int XrdDPMOssFile::Open(const char *fn, int Oflag, mode_t mode, XrdOucEnv &env) {

	const char *tident = "tident";
	EPNAME("XrdDPMOssFile::Open");
	XrdOucString surl,sfn;

	myOflag = Oflag;

	// get identity to use for dpm
	try {
		identP.reset(new DpmIdentity(&env));
        } catch(DpmIdentityException &e) {
		XrdOucString err =
			XrdOucString("Authentication error: ")
			+ e.DetailStr();
		OssEroute.Emsg("Open", SafeCStr(err));
  	        return -EINVAL;
        }

	// If supplied store the dpmtoken.
	// This is needed to invoke putdone
	char *tk = env.Get("dpm.tk");
	if (tk) {
		mydpmtk = strdup(tk);
	}


	// Get and store the surl that was used by the Finder
	// This is needed to invoke putdone
	try {
		surl = DecodeString(env.Get("dpm.surl"));
	} catch(const char *e) {
		return -EINVAL;
	}
	if (surl.length()) {
		myfilename = strdup(SafeCStr(surl));
	} else {
		myfilename = strdup(fn);
	}

	// Get the real sfn from the cgi info
	try {
		sfn = DecodeString(env.Get("dpm.sfn"));
	} catch(const char *e) {
		return -EINVAL;
	}
	if (!sfn.length()) {
		return -EINVAL;
	}

	// And then we forward the open to the builtin procedure, using the new sfn
	if (myfilename) {
		DEBUG("opening " << fn << " --> " << myfilename << " --> " << sfn);
	}
	int r = XrdOssFile::Open(SafeCStr(sfn), Oflag, 0660, env);

	if (r >= 0) {
		fListItem item(this,fn,sfn);
		XrdSysMutexHelper hlpMtx(&fLLock);
		fList.push_back(item);
	}

	return r;
}

int XrdDPMOssFile::Close(long long *retsz) {
	// Here we tell to the DPM daemon in the head node that we have finished
	// The dpm token is in mydpmtk
	int nbreplies, ret, ret_code = 0;
	struct dpm_filestatus *dpmfs = 0;

	const char *tident = "";
	EPNAME("XrdDPMOssFile::Close");
	DpmError dpmerr(epname);
	bool secSet = 1;

	try {
		if (!identP.get())
			throw DpmIdentityException(EINVAL, "No identity set");
		identP->SetAuth();
        } catch(DpmIdentityException &e) {
		secSet = 0;
        }


	if (myOflag & (O_WRONLY | O_RDWR)) {
		if (!secSet || !mydpmtk || !myfilename) {
			ret_code = -EINVAL;
		} else {
			if ((ret = dpm_putdone(mydpmtk, 1, &myfilename, &nbreplies, &dpmfs))) {
				dpmerr.save("dpm_putdone");
				ret_code = -dpmerr.GetSerrno();
			}
			DEBUG("dpm_putdone(" << mydpmtk << ", " << myfilename << ") returned " << ret);
			if (ret) {
				dpm_free_filest(nbreplies, dpmfs);
				dpmfs = 0;
				ret = dpm_abortfiles(mydpmtk, 1, &myfilename, &nbreplies, &dpmfs);
				DEBUG("dpm_abortfiles(" << mydpmtk << ", " << myfilename << ") returned " << ret);
			}
		}
	}
	dpm_free_filest(nbreplies, dpmfs);
	dpmfs = 0;

	if (secSet)
		identP->UnsetAuth();

	ret = XrdOssFile::Close(retsz);
	DEBUG("XrdOssFile::Close returned " << ret);

	if (ret >= 0)
		deleteObjFList(this);
	else
		ret_code = ret;

	DEBUG("XrdDPMOssFile::Close returning " << ret_code);
	return ret_code;
}

int XrdDPMOss::Chmod(const char *path, mode_t mode, XrdOucEnv *eP) {

	return -EINVAL;
}

int XrdDPMOss::Create(const char *tident, const char *path, mode_t access_mode,
	XrdOucEnv &env, int Opts) {

	return -ENOTSUP;
}

int XrdDPMOss::Init(XrdSysLogger *lp, const char *configfn) {
	// Just forward the initialization to the internal oss
	return (ossroute->Init(lp, configfn));
}

int XrdDPMOss::Mkdir(const char *path, mode_t mode, int mkpath, XrdOucEnv *envP) {

	return -EINVAL;
}

int XrdDPMOss::Remdir(const char *, int Opts, XrdOucEnv *eP) {

	return -EINVAL;
}

int XrdDPMOss::Rename(const char *, const char *, XrdOucEnv *eP1, XrdOucEnv *eP2) {

	return -EINVAL;
}

int XrdDPMOss::Stat(const char *fn, struct stat *st, int opts, XrdOucEnv *eP) {
	const char *tident = "";
	EPNAME("XrdDPMOss::Stat");

        if (!eP) {
                TRACE(ALL, "No environment parameters passed.");
                return -EINVAL;
        }

	memset(st, 0, sizeof (struct stat));
	
	// If this was requested for a Locate, then just say yes
	char *skip = eP->Get("dpm.skiplocate");
	if (skip && (*skip == '1')) {
		eP->Put("dpm.skiplocate", "0");
		return 0;
	}
	
	// Get the info from dpns
	struct Cns_filestat statbuf;
	memset(&statbuf, 0, sizeof (statbuf));

	// Xlate the fn
	XrdOucString myfn;

	try {
		myfn = DecodeString( eP->Get("dpm.surl"));
	} catch(const char *e) {
		return -EINVAL;
	}
	if (!myfn.length()) {
		myfn.assign(fn,0);
	}

	std::auto_ptr<DpmIdentity> identP;
	try {
		identP.reset(new DpmIdentity(eP));
                identP->SetAuth();
        } catch(DpmIdentityException &e) {
		XrdOucString err =
			XrdOucString("Authentication error: ")
			+ e.DetailStr();
		OssEroute.Emsg("Stat", SafeCStr(err));
                return -EINVAL;
        }

	// Do the stat
	DpmError dpmerr(epname);
	DEBUG("stat-ing " << myfn);
	if (! dpns_stat(SafeCStr(myfn), &statbuf)) {
		st->st_atim.tv_sec = statbuf.atime;
		st->st_ctim.tv_sec = statbuf.ctime;
		st->st_mode = statbuf.filemode;
		st->st_nlink = statbuf.nlink;
		st->st_mtim.tv_sec = statbuf.mtime;
		st->st_size = statbuf.filesize;
		st->st_ino = statbuf.fileid;
		return 0;
	}

	dpmerr.save("dpns_stat");
	return -dpmerr.GetSerrno();
}

int XrdDPMOss::Unlink(const char *fn, int Opts, XrdOucEnv *eP) {
	const char *tident = "";
	EPNAME("XrdDPMOss::Unlink");

	std::vector<XrdOucString> Sfns;

	getSfnsFromList(0, fn, Sfns);

	if (!Sfns.size())
		return -EINVAL;

	DEBUG("Returned vec of " << Sfns.size() << " first sfn=" << Sfns[0]);

	return ossroute->Unlink(SafeCStr(Sfns[0]), Opts, eP);
}

int XrdDPMOss::Lfn2Pfn(const char *Path, char *buff, int blen) {
	if ((int) strlen(Path) >= blen) return -ENAMETOOLONG;
	strcpy(buff, Path);
	return 0;
}

const char *XrdDPMOss::Lfn2Pfn(const char *Path, char *buff, int blen, int &rc) {
	rc = 0;
	return Path;
}


// -------------------------

XrdOss *XrdOssGetStorageSystem(XrdOss *native_oss,
	XrdSysLogger *Logger,
	const char *config_fn,
	const char *parms) {

	const char *tident = "";
	EPNAME("XrdOssGetStorageSystem");

	OssEroute.logger(Logger);
	OssEroute.Say(
			"This is XrdDPMOss "
			XrdDPMVERSIONSTR
			" compiled using xroot "
			XrdCompileVersion
		);

	XrdDPMOss *myoss = new XrdDPMOss(native_oss, (char *) parms);
	if (myoss) {
		
		myoss->Init(Logger, config_fn);
		TRACE(ALL, "Created DPM oss instance.");

		// Parse the parameters
		return myoss;
	}

	return 0;
}
